# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data uploader base class."""
import abc
from abc import ABC
import pandas as pd

from ..._version import VERSION

__version__ = VERSION
__author__ = "Pete Bryan"


class UploaderBase(ABC):
    """Base class for data providers."""

    def __init__(self, **kwargs):
        """Initialize new instance."""
        self._kwargs = kwargs
        self.workspace = None
        self.workspace_secret = None
        self._connected = False
        self._debug = False

    @abc.abstractmethod
    def upload_file(self, file_path: str, table_name: str, delim: str = ",", **kwargs):
        """
        Upload a file to the data store.

        Parameters
        ----------
        file_path : str
            Path to the file to upload
        table_name : str
            The name of the table to upload the file to
        delim : Optional[str]
            Column deliminator in data file, default is ,

        """

    @abc.abstractmethod
    def upload_folder(
        self, folder_path: str, table_name: str = None, delim: str = ",", **kwargs
    ):
        """
        Upload a folder of files to the data store.

        Parameters
        ----------
        folder_path : str
            Path to the folder of files to upload
        table_name : Optional[str]
            The name of the table to upload the file to, if not set file name is used as table name
        delim : Optional[str]
            Column deliminator in data file, default is ,

        """

    @abc.abstractmethod
    def upload_df(self, data: pd.DataFrame, table_name: str, **kwargs):
        """
        Upload a Pandas DataFrame to the data store.

        Parameters
        ----------
        data : pd.DataFrame
            The DataFrame to upload
        table_name : str
            The name of the table to upload the DataFrame to

        """



"""VirusTotal v3 API."""
from enum import Enum
from typing import Dict, List, Set

import pandas as pd
from IPython.display import HTML, display

from ..common.exceptions import MsticpyImportExtraError

try:
    import vt
    from vt_graph_api import VTGraph
    from vt_graph_api import errors as vt_graph_errs
except ImportError as imp_err:
    raise MsticpyImportExtraError(
        "Cannot use this feature without vt-py and vt-graph-api packages installed.",
        title="Error importing VirusTotal modules.",
        extra="vt3",
    ) from imp_err


class MsticpyVTNoDataError(Exception):
    """No data returned from VT API."""


class MsticpyVTGraphSaveGraphError(Exception):
    """Could not save VT Graph."""


class VTEntityType(Enum):
    """VTEntityType: Enum class for VirusTotal entity types."""

    FILE = "file"
    DOMAIN = "domain"
    IP_ADDRESS = "ip_address"
    URL = "url"


class ColumnNames(Enum):
    """Column name enum for DataFrame output."""

    ID = "id"
    TYPE = "type"
    DETECTIONS = "detections"
    SCANS = "scans"
    SOURCE = "source"
    TARGET = "target"
    RELATIONSHIP_TYPE = "relationship_type"
    SOURCE_TYPE = "source_type"
    TARGET_TYPE = "target_type"


class VTObjectProperties(Enum):
    """Enum for VT Object properties."""

    ATTRIBUTES = "attributes"
    RELATIONSHIPS = "relationship"
    LAST_ANALYSIS_STATS = "last_analysis_stats"
    MALICIOUS = "malicious"


class VTLookupV3:
    """VTLookupV3: VirusTotal lookup of IoC reports."""

    _SUPPORTED_VT_TYPES: Set[VTEntityType] = {
        VTEntityType.FILE,
        VTEntityType.URL,
        VTEntityType.IP_ADDRESS,
        VTEntityType.DOMAIN,
    }

    _MAPPING_TYPES_ENDPOINT: Dict[VTEntityType, str] = {
        VTEntityType.FILE: "files",
        VTEntityType.URL: "urls",
        VTEntityType.IP_ADDRESS: "ip_addresses",
        VTEntityType.DOMAIN: "domains",
    }

    _BASIC_PROPERTIES_PER_TYPE: Dict[VTEntityType, Set[str]] = {
        VTEntityType.FILE: {
            "type_description",
            "size",
            "first_submission_date",
            "last_submission_date",
            "times_submitted",
            "meaningful_name",
        },
        VTEntityType.URL: {
            "first_submission_date",
            "last_submission_date",
            "times_submitted",
        },
        VTEntityType.IP_ADDRESS: {"date", "country", "asn", "as_owner"},
        VTEntityType.DOMAIN: {"id", "creation_date", "last_update_date", "country"},
    }

    @property
    def supported_vt_types(self) -> List[str]:
        """
        Return list of VirusTotal supported IoC type names.

        Returns
        -------
        List[str]:
            List of VirusTotal supported IoC type names.

        """
        return [str(i_type) for i_type in self._SUPPORTED_VT_TYPES]

    @classmethod
    def _get_endpoint_name(cls, vt_type: str) -> str:
        if VTEntityType(vt_type) not in cls._SUPPORTED_VT_TYPES:
            raise KeyError(f"Property type {vt_type} not supported")

        return cls._MAPPING_TYPES_ENDPOINT[VTEntityType(vt_type)]

    @classmethod
    def _parse_vt_object(cls, vt_object: vt.object.Object) -> pd.DataFrame:
        obj_dict = vt_object.to_dict()
        if VTObjectProperties.ATTRIBUTES.value in obj_dict:
            attributes = obj_dict[VTObjectProperties.ATTRIBUTES.value]
            vt_type = VTEntityType(vt_object.type)
            if vt_type not in cls._SUPPORTED_VT_TYPES:
                raise KeyError(f"Property type {vt_type} not supported")
            obj = {
                key: attributes[key]
                for key in cls._BASIC_PROPERTIES_PER_TYPE[vt_type]
                if key in attributes
            }
            vt_df = pd.json_normalize(data=[obj])
            last_analysis_stats = attributes[
                VTObjectProperties.LAST_ANALYSIS_STATS.value
            ]
            vt_df[ColumnNames.DETECTIONS.value] = last_analysis_stats[
                VTObjectProperties.MALICIOUS.value
            ]
            vt_df[ColumnNames.SCANS.value] = sum(last_analysis_stats.values())
        else:
            vt_df = pd.DataFrame()

        # Inject ID and Type columns
        vt_df[ColumnNames.ID.value] = [vt_object.id]
        vt_df[ColumnNames.TYPE.value] = [vt_object.type]
        return vt_df.set_index([ColumnNames.ID.value])

    def __init__(self, vt_key: str):
        """
        Create a new instance of VTLookupV3 class.

        Parameters
        ----------
        vt_key: str
            VirusTotal API key

        """
        self._vt_key = vt_key
        self._vt_client = vt.Client(apikey=vt_key)

    def lookup_ioc(self, observable: str, vt_type: str) -> pd.DataFrame:
        """
        Look up and single IoC observable.

        Parameters
        ----------
        observable: str
            The observable value
        vt_type: str
            The VT entity type

        Returns
        -------
            Attributes Pandas DataFrame with the properties of the entity

        Raises
        ------
        KeyError
            Unknown vt_type

        """
        if VTEntityType(vt_type) not in self._SUPPORTED_VT_TYPES:
            raise KeyError(f"Property type {vt_type} not supported")

        endpoint_name = self._get_endpoint_name(vt_type)
        try:
            response: vt.object.Object = self._vt_client.get_object(
                f"/{endpoint_name}/{observable}"
            )
            return self._parse_vt_object(response)
        except vt.APIError as err:
            raise MsticpyVTNoDataError(
                "An error occurred requesting data from VirusTotal"
            ) from err
        finally:
            self._vt_client.close()

    def lookup_iocs(
        self,
        observables_df: pd.DataFrame,
        observable_column: str = ColumnNames.TARGET.value,
        observable_type_column: str = ColumnNames.TARGET_TYPE.value,
    ):
        """
        Look up and multiple IoC observables.

        Parameters
        ----------
        observables_df: pd.DataFrame
            A Pandas DataFrame, where each row is an observable
        observable_column:
            ID column of each observable
        observable_type_column:
            Type column of each observable

        Returns
        -------
            Attributes Pandas DataFrame with the properties of the entities

        Raises
        ------
        KeyError
            Column not found in observables_df

        """
        _observables_df = observables_df.reset_index()

        for column in [observable_column, observable_type_column]:
            if column not in _observables_df.columns:
                raise KeyError(f"Column {column} not found in observables_df")

        observables_list = _observables_df[observable_column]
        types_list = _observables_df[observable_type_column]
        dfs = []
        for observable, observable_type in zip(observables_list, types_list):
            try:
                ioc_df = self.lookup_ioc(observable, observable_type)
                dfs.append(ioc_df)
            except KeyError:
                print(
                    "ERROR\t It was not possible to obtain results for",
                    f"{observable_type} {observable}",
                )
                dfs.append(
                    pd.DataFrame(
                        data=[[observable, observable_type]],
                        columns=[ColumnNames.ID.value, ColumnNames.TYPE.value],
                    ).set_index(ColumnNames.ID.value)
                )
        return pd.concat(dfs) if dfs else pd.DataFrame()

    def lookup_ioc_relationships(
        self, observable: str, vt_type: str, relationship: str, limit: int = None
    ) -> pd.DataFrame:
        """
        Look up and single IoC observable relationships.

        Parameters
        ----------
        observable: str
            The observable value
        vt_type: str
            The VT entity type
        relationship: str
            Desired relationship
        limit: int
            Relations limit

        Returns
        -------
            Relationship Pandas DataFrame with the relationships of the entity

        Raises
        ------
        KeyError
            Unknown vt_type

        """
        if VTEntityType(vt_type) not in self._SUPPORTED_VT_TYPES:
            raise KeyError(f"Property type {vt_type} not supported")

        endpoint_name = self._get_endpoint_name(vt_type)
        response: vt.object.Object

        if limit is None:
            try:
                response = self._vt_client.get_object(
                    f"/{endpoint_name}/{observable}?relationship_counters=true"
                )
                relationships = response.relationships
                limit = (
                    relationships[relationship]["meta"]["count"]
                    if relationship in relationships
                    else 0
                )
            except KeyError:
                print(
                    f"ERROR: Could not obtain relationship limit for {vt_type} {observable}"
                )
                return pd.DataFrame()

        if limit == 0 or limit is None:
            return pd.DataFrame()

        try:
            # print(f"Obtaining {limit} relationships for {vt_type} {observable}")
            response = self._vt_client.iterator(
                f"/{endpoint_name}/{observable}/relationships/{relationship}",
                batch_size=40,
                limit=limit,
            )
            vt_objects = [self._parse_vt_object(r) for r in response]
            result_df = pd.concat(vt_objects) if vt_objects else pd.DataFrame()

            if vt_objects:
                # Inject source and target columns
                result_df[ColumnNames.SOURCE.value] = observable
                result_df[ColumnNames.SOURCE_TYPE.value] = VTEntityType(vt_type).value
                result_df[ColumnNames.RELATIONSHIP_TYPE.value] = relationship
                result_df.reset_index(inplace=True)
                result_df.rename(
                    columns={
                        ColumnNames.ID.value: ColumnNames.TARGET.value,
                        ColumnNames.TYPE.value: ColumnNames.TARGET_TYPE.value,
                    },
                    inplace=True,
                )
                result_df.set_index(
                    [ColumnNames.SOURCE.value, ColumnNames.TARGET.value], inplace=True
                )
        except vt.APIError as err:
            raise MsticpyVTNoDataError(
                "An error occurred requesting data from VirusTotal"
            ) from err

        finally:
            self._vt_client.close()

        return result_df

    def lookup_iocs_relationships(
        self,
        observables_df: pd.DataFrame,
        relationship: str,
        observable_column: str = ColumnNames.TARGET.value,
        observable_type_column: str = ColumnNames.TARGET_TYPE.value,
        limit: int = None,
    ) -> pd.DataFrame:
        """
        Look up and single IoC observable relationships.

        Parameters
        ----------
        observables_df: pd.DataFrame
            A Pandas DataFrame, where each row is an observable
        relationship: str
            Desired relationship
        observable_column:
            ID column of each observable
        observable_type_column:
            Type column of each observable.
        limit: int
            Relations limit

        Returns
        -------
            Relationship Pandas DataFrame with the relationships of each observable.

        Raises
        ------
        KeyError
            Column not found in observables_df

        """
        _observables_df = observables_df.reset_index()

        for column in [observable_column, observable_type_column]:
            if column not in _observables_df.columns:
                raise KeyError(f"Column {column} not found in observables df")

        observables_list = _observables_df[observable_column]
        types_list = _observables_df[observable_type_column]
        dfs = []

        for observable, observable_type in zip(observables_list, types_list):
            try:
                result_df = self.lookup_ioc_relationships(
                    observable, observable_type, relationship, limit
                )
                dfs.append(result_df)
            except KeyError:
                print(
                    "ERROR:\t It was not possible to get the data for",
                    f"{observable_type} {observable}",
                )
                dfs.append(
                    pd.DataFrame(
                        data=[[observable, observable_type]],
                        columns=[ColumnNames.ID.value, ColumnNames.TYPE.value],
                    ).set_index(ColumnNames.ID.value)
                )

        return pd.concat(dfs) if len(dfs) > 0 else pd.DataFrame()

    def create_vt_graph(
        self, relationship_dfs: List[pd.DataFrame], name: str, private: bool
    ) -> str:
        """
        Create a VirusTotal Graph with a set of Relationship DataFrames.

        Parameters
        ----------
        relationship_dfs:
            List of Relationship DataFrames
        name:
            New graph name
        private
            Indicates if the Graph is private or not.

        Returns
        -------
            Graph ID

        Raises
        ------
            ValueError when private is not indicated.
            ValueError when there are no relationship DataFrames
            MsticpyVTGraphSaveGraphError when Graph can not be saved

        """
        if len(relationship_dfs) == 0:
            raise ValueError("There are no relationship DataFrames")

        if not isinstance(private, bool):
            raise ValueError("Please indicate if Graph is private or not")

        concatenated_df = pd.concat(relationship_dfs).reset_index()

        # Create nodes DF, with source and target
        sources_df = (
            concatenated_df.groupby(ColumnNames.SOURCE.value)[
                ColumnNames.SOURCE_TYPE.value
            ]
            .first()
            .reset_index()
            .rename(
                columns={
                    ColumnNames.SOURCE.value: ColumnNames.ID.value,
                    ColumnNames.SOURCE_TYPE.value: ColumnNames.TYPE.value,
                }
            )
        )

        target_df = (
            concatenated_df.groupby(ColumnNames.TARGET.value)[
                ColumnNames.TARGET_TYPE.value
            ]
            .first()
            .reset_index()
            .rename(
                columns={
                    ColumnNames.TARGET.value: ColumnNames.ID.value,
                    ColumnNames.TARGET_TYPE.value: ColumnNames.TYPE.value,
                }
            )
        )

        nodes_df = pd.concat([sources_df, target_df])

        graph = VTGraph(self._vt_key, name=name, private=private)

        for _, row in nodes_df.iterrows():
            graph.add_node(
                node_id=row[ColumnNames.ID.value], node_type=row[ColumnNames.TYPE.value]
            )

        for _, row in concatenated_df.iterrows():
            graph.add_link(
                source_node=row[ColumnNames.SOURCE.value],
                target_node=row[ColumnNames.TARGET.value],
                connection_type=row[ColumnNames.RELATIONSHIP_TYPE.value],
            )
        try:
            graph.save_graph()
        except vt_graph_errs.SaveGraphError as graph_err:
            raise MsticpyVTGraphSaveGraphError(
                "Could not save Graph. %s" % ""
                if not private
                else "Please check you have Private Graph premium feature enabled in"
                "your subscription. It is possible to create public Graphs"
                "with 'private=False' input argument"
            ) from graph_err

        return graph.graph_id

    @staticmethod
    def render_vt_graph(graph_id: str, width: int = 800, height: int = 600):
        """
        Display a VTGraph in a Jupyter Notebook.

        Parameters
        ----------
        graph_id:
            Graph ID
        width
            Graph width.
        height
            Graph height

        """
        display(
            HTML(
                f"""
              <iframe
                src="https://www.virustotal.com/graph/embed/{graph_id}"
                width="{width}"
                height="{height}">
              </iframe>

            """
            )
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Helper module for computations when each session is a list of strings."""

from collections import defaultdict
from typing import Tuple, List, Union, DefaultDict
import copy

import numpy as np

from ..utils.data_structures import StateMatrix
from ..utils.laplace_smooth import laplace_smooth_cmd_counts
from ....common.exceptions import MsticpyException


def compute_counts(  # nosec
    sessions: List[List[str]], start_token: str, end_token: str, unk_token: str
) -> Tuple[DefaultDict[str, int], DefaultDict[str, DefaultDict[str, int]]]:
    """
    Compute counts of individual commands and of sequences of two commands.

    Parameters
    ----------
    sessions: List[List[str]]
        each session is a list of commands (strings)
        an example session:
            ['Set-User', 'Set-Mailbox']
    start_token: str
        dummy command to signify the start of a session (e.g. "##START##")
    end_token: str
        dummy command to signify the end of a session (e.g. "##END##")
    unk_token: str
        dummy command to signify an unseen command (e.g. "##UNK##")

    Returns
    -------
    tuple of counts:
        individual command counts,
        sequence command (length 2) counts

    """
    if not start_token != end_token != unk_token:
        raise MsticpyException(
            "start_token, end_token, unk_tokens should all be set to something "
            "different"
        )

    seq1_counts: DefaultDict[str, int] = defaultdict(lambda: 0)
    seq2_counts: DefaultDict[str, DefaultDict[str, int]] = defaultdict(
        lambda: defaultdict(lambda: 0)
    )

    for session in sessions:
        prev = start_token
        seq1_counts[prev] += 1
        for cmd in session:
            seq1_counts[cmd] += 1
            seq2_counts[prev][cmd] += 1
            prev = cmd
        seq2_counts[prev][end_token] += 1
        seq1_counts[end_token] += 1

    return seq1_counts, seq2_counts


def laplace_smooth_counts(
    seq1_counts: DefaultDict[str, int],
    seq2_counts: DefaultDict[str, DefaultDict[str, int]],
    start_token: str,
    end_token: str,
    unk_token: str,
) -> Tuple[StateMatrix, StateMatrix]:
    """
    Laplace smoothing is applied to the counts.

    We do this by adding 1 to each of the counts. This is so when we
    compute the probabilities from the counts, we shift some of the
    probability mass from the very probable commands and command sequences
    to the unseen and very unlikely commands and command sequences.
    The `unk_token` means we can handle unseen commands and sequences of commands.

    Parameters
    ----------
    seq1_counts: DefaultDict[str, int]
        individual command counts
    seq2_counts: DefaultDict[str, DefaultDict[str, int]]
        sequence command (length 2) counts
    start_token: str
        dummy command to signify the start of a session (e.g. "##START##")
    end_token: str
        dummy command to signify the end of a session (e.g. "##END##")
    unk_token: str
        dummy command to signify an unseen command (e.g. "##UNK##")

    Returns
    -------
    tuple of StateMatrix laplace smoothed counts:
        individual command counts,
        sequence command (length 2) counts

    """
    # apply laplace smoothing
    seq1_counts_ls, seq2_counts_ls = laplace_smooth_cmd_counts(
        seq1_counts=copy.deepcopy(seq1_counts),
        seq2_counts=copy.deepcopy(seq2_counts),
        start_token=start_token,
        end_token=end_token,
        unk_token=unk_token,
    )

    seq1_counts_st = StateMatrix(states=seq1_counts_ls, unk_token=unk_token)
    seq2_counts_st = StateMatrix(states=seq2_counts_ls, unk_token=unk_token)

    return seq1_counts_st, seq2_counts_st


# pylint: disable=too-many-arguments, too-many-branches
def compute_likelihood_window(
    window: List[str],
    prior_probs: Union[StateMatrix, dict],
    trans_probs: Union[StateMatrix, dict],
    use_start_token: bool,
    use_end_token: bool,
    start_token: str = None,
    end_token: str = None,
) -> float:
    """
    Compute the likelihood of the input `window`.

    Parameters
    ----------
    window: List[str]
        part or all of a session, where a session is a list of commands (strings)
        an example session:
            ['Set-User', 'Set-Mailbox']
    prior_probs: Union[StateMatrix, dict]
        computed probabilities of individual commands
    trans_probs: Union[StateMatrix, dict]
        computed probabilities of sequences of commands (length 2)
    use_start_token: bool
        if set to True, the start_token will be prepended to the window
        before the likelihood calculation is done
    use_end_token: bool
        if set to True, the end_token will be appended to the window
        before the likelihood calculation is done
    start_token: str
        dummy command to signify the start of the session (e.g. "##START##")
    end_token: str
        dummy command to signify the end of the session (e.g. "##END##")

    Returns
    -------
    likelihood of the window

    """
    if use_start_token:
        if start_token is None:
            raise MsticpyException(
                "start_token should not be None, when use_start_token is True"
            )

    if use_end_token:
        if end_token is None:
            raise MsticpyException(
                "end_token should not be None, when use_end_token is True"
            )

    w_len = len(window)
    if w_len == 0:
        return np.nan
    prob = 1

    cur = window[0]
    if use_start_token:
        prob *= trans_probs[start_token][cur]
    else:
        prob *= prior_probs[cur]

    for i in range(1, w_len):
        prev, cur = window[i - 1], window[i]
        prob *= trans_probs[prev][cur]

    if use_end_token:
        prob *= trans_probs[cur][end_token]

    return prob


# pylint: disable=too-many-locals, too-many-arguments, too-many-branches
# pylint: disable=too-many-locals, too-many-branches
def compute_likelihood_windows_in_session(
    session: List[str],
    prior_probs: Union[StateMatrix, dict],
    trans_probs: Union[StateMatrix, dict],
    window_len: int,
    use_start_end_tokens: bool,
    start_token: str = None,
    end_token: str = None,
    use_geo_mean: bool = False,
) -> List[float]:
    """
    Compute the likelihoods of a sliding window of length `window_len` in the session.

    Parameters
    ----------
    session: List[str]
        list of commands (strings)
        an example session:
            ['Set-User', 'Set-Mailbox']
    prior_probs: Union[StateMatrix, dict]
        computed probabilities of individual commands
    trans_probs: Union[StateMatrix, dict]
        computed probabilities of sequences of commands (length 2)
    window_len: int
        length of sliding window for likelihood calculations
    use_start_end_tokens: bool
        if True, then `start_token` and `end_token` will be prepended and appended to the
        session respectively before the calculations are done
    start_token: str
        dummy command to signify the start of the session (e.g. "##START##")
    end_token: str
        dummy command to signify the end of the session (e.g. "##END##")
    use_geo_mean: bool
        if True, then each of the likelihoods of the sliding windows will be
        raised to the power of (1/`window_len`)

    Returns
    -------
    list of likelihoods

    """
    if use_start_end_tokens:
        if start_token is None or end_token is None:
            raise MsticpyException(
                "start_token and end_token should not be set to None when "
                "use_start_end_tokens is set to True"
            )

    likelihoods = []
    sess = session.copy()
    if use_start_end_tokens and end_token:
        sess += [str(end_token)]
    end = len(sess) - window_len
    for i in range(end + 1):
        window = sess[i : i + window_len]  # noqa: E203

        if i == 0:
            use_start = use_start_end_tokens
        else:
            use_start = False

        lik = compute_likelihood_window(
            window=window,
            prior_probs=prior_probs,
            trans_probs=trans_probs,
            use_start_token=use_start,
            use_end_token=False,
            start_token=start_token,
            end_token=end_token,
        )
        if use_geo_mean:
            k = window_len
            lik = lik ** (1 / k)

        likelihoods.append(lik)

    return likelihoods


# pylint: disable=too-many-arguments
def rarest_window_session(
    session: List[str],
    prior_probs: Union[StateMatrix, dict],
    trans_probs: Union[StateMatrix, dict],
    window_len: int,
    use_start_end_tokens: bool,
    start_token: str,
    end_token: str,
    use_geo_mean: bool = False,
) -> Tuple[List[str], float]:
    """
    Find and compute likelihood of the rarest window in the session.

    Parameters
    ----------
    session: List[str]
        list of commands (strings)
        an example session:
            ['Set-User', 'Set-Mailbox']
    prior_probs: Union[StateMatrix, dict]
        computed probabilities of individual commands
    trans_probs: Union[StateMatrix, dict]
        computed probabilities of sequences of commands (length 2)
    window_len: int
        length of sliding window for likelihood calculations
    use_start_end_tokens: bool
        if True, then `start_token` and `end_token` will be prepended
        and appended to the session respectively before the calculations are done
    start_token: str
        dummy command to signify the start of the session (e.g. "##START##")
    end_token: str
        dummy command to signify the end of the session (e.g. "##END##")
    use_geo_mean: bool
        if True, then each of the likelihoods of the sliding windows will be
        raised to the power of (1/`window_len`)

    Returns
    -------
    (rarest window part of the session, likelihood of the rarest window)

    """
    likelihoods = compute_likelihood_windows_in_session(
        session=session,
        prior_probs=prior_probs,
        trans_probs=trans_probs,
        window_len=window_len,
        use_start_end_tokens=use_start_end_tokens,
        start_token=start_token,
        end_token=end_token,
        use_geo_mean=use_geo_mean,
    )

    if len(likelihoods) == 0:
        return [], np.nan
    min_lik = min(likelihoods)
    ind = likelihoods.index(min_lik)
    return session[ind : ind + window_len], min_lik  # noqa: E203



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Module for VTLookup class.

Wrapper class around `Virus Total
API <https://www.virustotal.com/en/documentation/public-api/>`__. Input
can be a single IoC observable or a pandas DataFrame containing multiple
observables. Processing requires a Virus Total account and API key and
processing performance is limited to the number of requests per minute
for the account type that you have. Support IoC Types:

-  Filehash
-  URL
-  DNS Domain
-  IPv4 Address

"""
# pylint: disable=too-many-lines
import json
from json import JSONDecodeError
from typing import List, Mapping, Any, Dict, Optional, Tuple
from collections import namedtuple

import pandas as pd
import requests

from .iocextract import IoCExtract
from .tiproviders.ti_provider_base import SanitizedObservable, preprocess_observable
from ..common.utility import export
from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"

# VirusTotal parameter collection
VTParams = namedtuple(
    "VTParams",
    [
        "api_type",
        "batch_size",
        "batch_delimiter",
        "http_verb",
        "api_var_name",
        "headers",
    ],
)

DuplicateStatus = namedtuple("DuplicateStatus", ["is_dup", "status"])


@export
class VTLookup:
    """
    VTLookup: VirusTotal lookup of IoC reports.

    Main methods are:
    lookup_iocs() - accepts input of multiple IoCs in a Pandas DataFrame
    lookup_ioc() - looks up a single IoC observable.
    supported_ioc_types - a list of valid target types.
    ioc_vt_type_mapping - a dictionary of mappings to recognized VT Types.
    Types mapped to None will not be submitted to VT.

    For urls a full http request can be submitted, query string and fragments will be
    dropped before submitting.
    For files MD5, SHA1 and SHA256 hashes are supported.
    For IP addresses only dotted IPv4 addresses are supported.
    """

    # Ioc types that we support
    _SUPPORTED_INPUT_TYPES: List[str] = [
        "ipv4",
        "dns",
        "url",
        "md5_hash",
        "sha1_hash",
        "sh256_hash",
    ]

    # Mapping to to VT Types
    _VT_TYPE_MAP: Dict[str, str] = {
        "ipv4": "ip-address",
        "dns": "domain",
        "url": "url",
        "md5_hash": "file",
        "sha1_hash": "file",
        "sh256_hash": "file",
    }

    # VT API parameters
    _HDR_GZIP = {"Accept-Encoding": "gzip, deflate"}
    _VT_API = "https://www.virustotal.com/vtapi/v2/{type}/report"
    _VT_API_TYPES: Dict[str, VTParams] = {
        "url": VTParams("url", 1, "\n", "get", "resource", _HDR_GZIP),
        "file": VTParams("file", 25, ",", "get", "resource", _HDR_GZIP),
        "ip-address": VTParams("ip-address", 1, "", "get", "ip", None),
        "domain": VTParams("domain", 1, "", "get", "domain", None),
    }

    _RESULT_COLUMNS: List[str] = [
        "Observable",
        "IoCType",
        "Status",
        "ResponseCode",
        "RawResponse",
        "Resource",
        "SourceIndex",
        "VerboseMsg",
        "Resource",
        "ScanId",
        "Permalink",
        "Positives",
        "MD5",
        "SHA1",
        "SHA256",
        "ResolvedDomains",
        "ResolvedIPs",
        "DetectedUrls",
    ]

    _http_strict_rgxc = None  # type: Any

    def __init__(self, vtkey: str, verbosity: int = 1):
        """
        Create a new instance of VTLookup class.

        Parameters
        ----------
        vtkey : str
            VirusTotal API key
        verbosity : int, optional
            The level of detail of reporting
                0 = no reporting
                1 = minimal reporting (default)
                2 = verbose reporting

        """
        self._vtkey = vtkey
        self._verbosity = verbosity
        self._ioc_custom_type_map = {}  # type: Dict[str, Optional[str]]

        self._ioc_extract = IoCExtract()

        # create a data frame to store the results
        self.results = pd.DataFrame(data=None, columns=self._RESULT_COLUMNS)

    @property
    def supported_ioc_types(self) -> List[str]:
        """
        Return list of supported IoC type internal names.

        Returns
        -------
        List[str]
            List of supported IoC type internal names.

        """
        return self._SUPPORTED_INPUT_TYPES

    @property
    def supported_vt_types(self) -> List[str]:
        """
        Return list of VirusTotal supported IoC type names.

        Returns
        -------
        List[str]
            List of VirusTotal supported IoC type names.

        """
        return list(self._VT_API_TYPES.keys())

    @property
    def ioc_vt_type_mapping(self) -> Dict[str, str]:
        """
        Return mapping between internal and VirusTotal IoC type names.

        Returns
        -------
        Mapping[str, str]
            Return mapping between internal and VirusTotal IoC type names.

        """
        return self._VT_TYPE_MAP

    def lookup_iocs(
        self,
        data: pd.DataFrame,
        src_col: str = "Observable",
        type_col: str = "IoCType",
        src_index_col: str = "SourceIndex",
        **kwargs,
    ) -> pd.DataFrame:
        """
        Retrieve results for IoC observables in the source dataframe.

        Parameters
        ----------
        data : pd.DataFrame
            Dataframe containing the observables to search for
        src_col : str, optional
            The column name that contains the observable data
            (one item per row) (the default is 'Observable')
        type_col : str, optional
            The column name containing the observable type
            (the default is 'IoCType')
        src_index_col : str, optional
            The name of the column to use as source index. If not
            specified this defaults to 'SourceIndex'. If this (or the supplied value)
            is not in the source dataframe, the index of the source dataframe will
            be used. This is retained in the output so that you can join the results
            back to the original data.
            (the default is 'SourceIndex')

        Other Parameters
        ----------------
            key/value pairs of additional mappings to supported IoC type names
            e.g. ipv4='ipaddress', url='httprequest'.
            This allows you to specify custom
            mappings when the source data is tagged with different names.

        Returns
        -------
        pd.DataFrame
            Combined results of local pre-processing and VirusTotal Lookups

        Raises
        ------
        KeyError
            Unknown ioc_type

        Notes
        -----
            See supported_ioc_types attribute for a list of valid target types.
            Not all of these types are supported by VirusTotal.
            See ioc_vt_type_mapping for current mappings.
            Types mapped to None will not be submitted to VT.

            For urls a full http request can be submitted, query string
            and fragments will be dropped before submitting.
            Other supported protocols are ftp, telnet, ldap, file
            For files MD5, SHA1 and SHA256 hashes are supported.
            For IP addresses only dotted IPv4 addresses are supported.

        """
        # if the caller has supplied alternative type name mappings add any of these
        # to our lookup dictionary
        for k in self._get_supported_vt_ioc_types():
            self._ioc_custom_type_map[k] = k
        for k, val in kwargs.items():
            if k in self._get_supported_vt_ioc_types():
                self._ioc_custom_type_map[k] = val

        src_idx_col = src_index_col if src_index_col in data else None

        # for each ioc_type, retrieve observables from dataframe
        for ioc_type, mapped_type in self._ioc_custom_type_map.items():
            input_df = data[data[type_col] == mapped_type]
            self._lookup_ioc_type(input_df, ioc_type, src_col, src_idx_col)

        self._print_status(
            "Submission complete. {} responses from {} input rows".format(
                len(self.results), len(data)
            ),
            2,
        )

        return self.results

    def lookup_ioc(self, observable: str, ioc_type: str, output: str = "dict") -> Any:
        """
        Look up and single IoC observable.

        Parameters
        ----------
        observable : str
            The observable value
        ioc_type : str
            The IoC Type (see 'supported_ioc_types' attribute)
        output : str, optional
            Output results as a dictionary (or list of dicts)
            if `output` is any other value the result will be returned in a
            Pandas DataFrame (the default is 'dict')

        Returns
        -------
            list{dict}: if output == 'dict'
            pd.DataFrame: otherwise

        Raises
        ------
        KeyError
            Unknown ioc_type

        """
        # Check input
        if (
            observable is None
            or observable.strip() is None
            or ioc_type is None
            or ioc_type.strip() is None
        ):
            raise SyntaxError("Invalid value for observable or ioc_type")

        observable, status = preprocess_observable(observable, ioc_type)
        if observable is None:
            raise SyntaxError("{} for observable value {}".format(status, observable))

        if ioc_type not in self._VT_TYPE_MAP:
            raise LookupError(
                "IoC Type {} not recognized. Valid types are [{}]".format(
                    ioc_type, ", ".join(self.supported_ioc_types)
                )
            )

        if self._VT_TYPE_MAP[ioc_type] not in self._VT_API_TYPES:
            vt_types = {
                k for k, val in self.ioc_vt_type_mapping.items() if val is not None
            }
            err = (
                "IoC Type {} is recognized by VirusTotal. Valid types are [{}]".format(
                    ioc_type, ", ".join(vt_types)
                )
            )
            raise LookupError(err)

        # do the submission
        vt_api_type = self._VT_TYPE_MAP[ioc_type]
        vt_param = self._VT_API_TYPES[vt_api_type]
        results, _ = self._vt_submit_request(observable, vt_param)
        self._parse_vt_results(results, observable, ioc_type)

        # return as a list of dictionaries or a DataFrame
        if output == "dict":
            list_res = self.results.apply(lambda x: x.to_dict(), axis=1).tolist()
            return list_res[0] if len(list_res) == 1 else list_res

        return self.results

    # pylint: disable=too-many-locals
    def _lookup_ioc_type(
        self,
        input_frame: pd.DataFrame,
        ioc_type: str,
        src_col: str,
        src_index_col: Optional[str],
    ):
        """
        Perform the VT submission of a set of IoCs of a given type.

        Parameters
        ----------
        input_frame : pd.DataFrame
            the input dataframe
        ioc_type : str
            the IoC Type to submit
        src_col : str
            The name column in the dataframe containing the
            IoC observables
        src_index_col : Optional[str]
            SourceIndex column name

        Raises
        ------
        KeyError
            Unknown ioc_type

        """
        if ioc_type not in self._VT_TYPE_MAP:
            raise KeyError(f'Unknown ioc_type "{ioc_type}""')

        vt_param = self._VT_API_TYPES[self._VT_TYPE_MAP[ioc_type]]

        # Some types support batch lookups so we can assemble them into batches
        # for the moment we are only supporting
        source_row_index = {}
        obs_batch = []
        batch_index = 0
        row_num = 0
        row_count = len(input_frame)
        if src_index_col:
            src_cols = [src_col, src_index_col]
        else:
            src_cols = [src_col]

        for idx, row in input_frame[src_cols].iterrows():
            row_num += 1
            observable = row[src_col]

            # Use the user-specified index if possible
            if src_index_col:
                idx = row[src_index_col]

            # validate the observable to avoid sending too much junk to VT
            pp_observable = self._validate_observable(observable, ioc_type, idx)

            # if the observable is valid, add it to the submission batch
            if pp_observable.observable:
                obs_batch.append(pp_observable.observable)
                source_row_index[pp_observable.observable] = idx
                batch_index += 1

            # We want to trigger in the following circumstances
            # 1. if the length of our batch is at the max VT batchsize for
            # this type (If the batch size is 1 this will fire for every row)
            # 2. Or we have reached the end of our row iteration
            # AND
            # 3. The batch is not empty
            if (
                len(obs_batch) == vt_param.batch_size or row_num == row_count
            ) and obs_batch:
                obs_submit = vt_param.batch_delimiter.join(obs_batch)

                self._print_status(
                    (
                        "Submitting observables: "
                        + f'"{obs_submit}", type "{ioc_type}" '
                        + "to VT. (Source index {idx})"
                    ),
                    2,
                )
                # Submit the request
                results, status_code = self._vt_submit_request(obs_submit, vt_param)

                if status_code != 200:
                    # Print status messages and add failure cases to results
                    status = "Failed submission: http error {}".format(status_code)
                    for failed_obs in obs_batch:
                        self._add_invalid_input_result(
                            failed_obs, ioc_type, status, source_row_index[failed_obs]
                        )
                        self._print_status(
                            'Error in response submitting observables: "{}", type "{}" '
                            "http status is {}. Response: {} (Source index {})".format(
                                obs_submit,
                                ioc_type,
                                status_code,
                                results,
                                source_row_index[failed_obs],
                            ),
                            1,
                        )
                else:
                    # parse the results from the response
                    self._parse_vt_results(
                        results, obs_submit, ioc_type, idx, source_row_index, vt_param
                    )

                # reset index of batch
                batch_index = 0
                obs_batch = []

    # pylint: disable=too-many-arguments, too-many-branches
    def _parse_vt_results(  # noqa: C901
        self,
        vt_results: Any,
        observable: str,
        ioc_type: str,
        source_idx: Any = 0,
        source_row_index: Any = None,
        vt_param: VTParams = None,
    ):
        """
        Parse VirusTotal results based on IoCType.

            :param vt_results: Raw results from VT
            :param observable: The observable or observable batch
            :param ioc_type: The IoC type of the observables
            :param source_idx: The row index of the source frame
            :param source_row_index: (batch only) Mapping between observable item
                and row index of the source
            :param vt_param: (batch only) the VTParams tuple for this submission

        """
        results_to_parse = []  # type: List[dict]
        if isinstance(vt_results, str):
            try:
                vt_results = json.loads(vt_results, strict=False)
            except (JSONDecodeError, TypeError):
                pass

        if (
            isinstance(vt_results, list)
            and vt_param is not None
            and vt_param.batch_size > 1
        ):
            # multiple results
            results_to_parse = vt_results
        elif isinstance(vt_results, dict):
            # single result
            results_to_parse.append(vt_results)
        else:
            self._print_status(
                (
                    "Error parsing response to JSON: "
                    + f'"{observable}", type "{ioc_type}". '
                    + f"(Source index {source_idx})"
                ),
                1,
            )

        if vt_param and vt_param.batch_delimiter:
            observables = observable.split(vt_param.batch_delimiter)
        else:
            observables = [observable]

        # pylint: disable=locally-disabled, consider-using-enumerate
        for result_idx in range(0, len(results_to_parse)):
            df_dict_vtresults = self._parse_single_result(
                results_to_parse[result_idx], ioc_type
            )

            # Add remaining fields from source
            df_dict_vtresults["IoCType"] = ioc_type
            df_dict_vtresults["Status"] = "Success"
            df_dict_vtresults["RawResponse"] = json.dumps(results_to_parse[result_idx])
            if (
                len(results_to_parse) == 1
                or source_row_index is None
                or len(source_row_index) == 1
            ):
                df_dict_vtresults["Observable"] = observable
                df_dict_vtresults["SourceIndex"] = source_idx
            else:
                # If we submitted multiple values in a batch
                # we assume (hope) that the ordering of the response is the same
                # as in the request. We try our best to remarry the observable
                # and source index
                if "resource" in results_to_parse[result_idx]:
                    vt_resource = results_to_parse[result_idx]["resource"]
                    df_dict_vtresults["Observable"] = vt_resource
                    if vt_resource in source_row_index:
                        df_dict_vtresults["SourceIndex"] = source_row_index[vt_resource]
                    else:
                        df_dict_vtresults["SourceIndex"] = source_row_index[
                            observables[result_idx]
                        ]
                else:
                    df_dict_vtresults["Observable"] = observables[result_idx]
                    df_dict_vtresults["SourceIndex"] = source_row_index[
                        observables[result_idx]
                    ]

            new_results = pd.concat(
                objs=[self.results, df_dict_vtresults], ignore_index=True, axis=0
            )

            self.results = new_results
        # pylint enable=locally-disabled, C0200

    def _parse_single_result(
        self, results_dict: Mapping[str, Any], ioc_type: str
    ) -> pd.DataFrame:
        """
        Parse VirusTotal single result based on IoCType.

        Parameters
        ----------
        results_dict : Mapping[str, Any]
            Raw results dictionary from VT
        ioc_type : str
            The IoC type of the observables

        Returns
        -------
        pd.DataFrame
            The results DataFrame

        """
        # create output frame and parse results to intermediate frame
        df_dict_vtresults = {}

        # Parse returned results to our output dataframe depending
        # on the IoC type
        if ioc_type in ["url", "md5_hash", "sha1_hash", "sha256_hash"]:
            df_dict_vtresults["ResponseCode"] = results_dict.get("response_code", None)
            df_dict_vtresults["VerboseMsg"] = results_dict.get("verbose_msg", None)
            df_dict_vtresults["ScanId"] = results_dict.get("scan_id", None)
            df_dict_vtresults["Resource"] = results_dict.get("resource", None)
            df_dict_vtresults["Permalink"] = results_dict.get("permalink", None)
            df_dict_vtresults["Positives"] = results_dict.get("positives", None)
            if ioc_type in ["md5_hash", "sha1_hash", "sha256_hash"]:
                df_dict_vtresults["MD5"] = results_dict.get("md5", None)
                df_dict_vtresults["SHA1"] = results_dict.get("sha1", None)
                df_dict_vtresults["SHA256"] = results_dict.get("sha256", None)

        if ioc_type in ["ipv4", "dns"]:
            df_dict_vtresults["ResponseCode"] = results_dict.get("response_code", None)
            df_dict_vtresults["VerboseMsg"] = results_dict.get("verbose_msg", None)
            # dns and ipv4 have multi-valued 'resolutions' and 'detected_urls' lists
            # of dictionaries
            # This leads to a few horrendous-looking list comprehensions
            # These are essentially pulling out the columns that contain these lists.
            # then using a list comprehension to pull out the value, where the key 'k'
            # is of the required value
            if ioc_type == "ipv4" and "resolutions" in results_dict:
                item_list = [
                    item["hostname"]
                    for item in results_dict["resolutions"]
                    if "hostname" in item
                ]
                df_dict_vtresults["ResolvedDomains"] = ", ".join(item_list)
            elif ioc_type == "dns" and "resolutions" in results_dict:
                item_list = [
                    item["ip_address"]
                    for item in results_dict["resolutions"]
                    if "ip_address" in item
                ]
                df_dict_vtresults["ResolvedIPs"] = ", ".join(item_list)
            if "detected_urls" in results_dict:
                item_list = [
                    item["url"]
                    for item in results_dict["detected_urls"]
                    if "url" in item
                ]
                df_dict_vtresults["DetectedUrls"] = ", ".join(item_list)
                # positives are listed per detected_url so we need to
                # pull those our and sum them.
                positives = sum(
                    [
                        item["positives"]
                        for item in results_dict["detected_urls"]
                        if "positives" in item
                    ]
                )
                df_dict_vtresults["Positives"] = positives

        return pd.DataFrame(
            data=df_dict_vtresults, columns=self._RESULT_COLUMNS, index=[0]
        )

    def _validate_observable(
        self, observable: str, ioc_type: str, idx: Any
    ) -> SanitizedObservable:
        """
        Validate observable for format and duplicates of existing results.

        Parameters
        ----------
        observable : str
            The observable to be checked
        ioc_type : str
            The IoCType of the observable
        idx : Any
            The index of the source row

        Returns
        -------
        SanitizedObservable
            The Pre-processed result

        """
        if observable is None or observable.strip() is None:
            status = "Failed: Empty or missing observable value"
            self._add_invalid_input_result(observable, ioc_type, status, idx)
            self._print_status(status + " (Source index {})".format(idx), 1)
            return SanitizedObservable(None, status)

        # Check that observable is of the correct format for this type
        # and do any cleaning up required
        pp_observable = preprocess_observable(observable, ioc_type)
        if pp_observable.observable is None:
            self._add_invalid_input_result(
                observable, ioc_type, pp_observable.status, idx
            )
            # pylint: disable=locally-disabled, line-too-long
            self._print_status(
                (
                    f'Invalid observable format: "{observable}", '
                    + f'type "{ioc_type}", '
                    + f"status: {pp_observable.status} "
                    + f"- skipping. (Source index {idx})"
                ),
                2,
            )
            # pylint: enable=locally-disabled, line-too-long
            return pp_observable

        # Check that we don't already have a result for this
        dup_result = self._check_duplicate_submission(observable, ioc_type, idx)
        if dup_result.is_dup:
            # pylint: disable=locally-disabled, line-too-long
            self._print_status(
                (
                    "Duplicate observable value detected: "
                    + f'"{observable}", type "{ioc_type}" '
                    + f"status: {dup_result.status} "
                    + f"- skipping. (Source index {idx})"
                ),
                2,
            )
            return SanitizedObservable(None, dup_result.status)

        return pp_observable

    def _check_duplicate_submission(
        self, observable: str, ioc_type: str, source_index: Any
    ) -> DuplicateStatus:
        """
        Check for a duplicate value in existing results.

        Parameters
        ----------
        observable : str
             The IoC observable value
        ioc_type : str
            The IoC type
        source_index : Any
            The index of the source DataFrame row

        Returns
        -------
        DuplicateStatus
            Status indicating whether this is a duplicate.

        """
        if self.results is None:
            return DuplicateStatus(False, "ok")

        # Note duplicate var here can be multiple rows of past results
        duplicate = self.results[self.results["Observable"] == observable].copy()
        # if this is a file hash we should check for previous results in
        # all of the hash columns
        if duplicate.shape[0] == 0 and ioc_type in [
            "md5_hash",
            "sha1_hash",
            "sh256_hash",
        ]:
            dup_query = (
                "MD5 == @observable or SHA1 == @observable or SHA256 == @observable"
            )
            duplicate = self.results.query(dup_query).copy()
            # In these cases we want to set the observable to the source value
            # but keep the rest of the results
            if duplicate.shape[0] > 0:
                duplicate["Observable"] = observable

        # if we found a duplicate so add the copies of the duplicated requests
        # to the results
        if duplicate.shape[0] > 0:
            original_indices = [v[0] for v in duplicate[["SourceIndex"]].values]
            duplicate["SourceIndex"] = source_index
            duplicate["Status"] = "Duplicate"
            new_results = pd.concat(
                objs=[self.results, duplicate], ignore_index=True, sort=False, axis=0
            )
            self.results = new_results

            return DuplicateStatus(True, "Duplicates of {}".format(original_indices))

        return DuplicateStatus(False, "ok")

    def _add_invalid_input_result(
        self, observable: str, ioc_type: str, status: str, source_idx: Any
    ):
        """
        Add a result row to indicate an invalid submission.

        Parameters
        ----------
        observable : str
            The IoC observable value
        ioc_type : str
            The IoC type
        status : str
            The status - why the item was invalid
        source_idx : Any
            The index of the source DataFrame row

        """
        new_row = pd.Series(index=self._RESULT_COLUMNS)
        new_row["Observable"] = observable
        new_row["IoCType"] = ioc_type
        new_row["Status"] = status
        new_row["SourceIndex"] = source_idx
        new_results = self.results.append(new_row.to_dict(), ignore_index=True)

        self.results = new_results

    def _vt_submit_request(
        self, submission_string: str, vt_param: VTParams
    ) -> Tuple[Optional[Dict[Any, Any]], int]:
        """
        Submit the request to VT.

        Parameters
        ----------
        submission_string : str
            The observable (or observable collection)
        vt_param : VTParams
            VT parameters appropriate to this observable type

        """
        params = {"apikey": self._vtkey, vt_param.api_var_name: submission_string}
        submit_url = self._get_vt_api_url(vt_param.api_type)
        headers = {"User-Agent": "VirusTotal", "Content-Type": "application/json"}
        if vt_param.headers is not None:
            for hdr, val in vt_param.headers.items():
                headers[hdr] = val

        if vt_param.http_verb == "post":
            response = requests.post(submit_url, data=params, headers=headers)
        else:
            response = requests.get(submit_url, params=params, headers=headers)
        if response.status_code == 200:
            return response.json(), response.status_code

        if response:
            try:
                return response.json(), response.status_code
            except JSONDecodeError:
                pass
        return None, response.status_code

    @classmethod
    def _get_vt_api_url(cls, api_type: str) -> str:
        """
        Return the VirusTotal API URL for the supplied type.

            :param api_type: The IoC type
        """
        if api_type not in cls._VT_API_TYPES:
            raise LookupError('Unknown api type "{}"'.format(api_type))
        return cls._VT_API.format(type=api_type)

    @classmethod
    def _get_supported_vt_ioc_types(cls) -> List[str]:
        """Return the subset of IoC types supported by VT."""
        return [
            t for t in cls._SUPPORTED_INPUT_TYPES if cls._VT_TYPE_MAP[t] is not None
        ]

    def _print_status(self, message: str, verbosity_level: int):
        """
        Print a status message depending on the current level of verbosity.

        Parameters
        ----------
        message : str
            the string message to print
        verbosity_level : int
            verbosity_level at which level the message should be output

        """
        if verbosity_level <= self._verbosity:
            print(message)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Initialization for Jupyter Notebooks."""
import os
from functools import wraps
import importlib
import sys
import warnings
from pathlib import Path
from typing import Any, List, Optional, Tuple, Dict, Callable

from IPython.core.interactiveshell import InteractiveShell
from IPython.display import display, HTML
import ipywidgets as widgets
from matplotlib import MatplotlibDeprecationWarning
import pandas as pd
import seaborn as sns

from ..common.exceptions import MsticpyUserError, MsticpyException
from ..common.utility import check_and_install_missing_packages, unit_testing, md
from ..common.pkg_config import validate_config, get_config
from ..common.wsconfig import WorkspaceConfig
from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


_IMPORT_ERR_MSSG = """
<h2><font color='red'>One or more missing packages detected</h2>
Please correct these by installing the required packages, restart
the kernel and re-run the notebook.</font>
<i>Package error: {err}</i><br>
"""

_IMPORT_MODULE_MSSG = """
<font color='red'>Error import module {module}</font>
"""

_MISSING_PKG_WARN = """
<h3><font color='orange'>Warning {package} is not installed or has an
incorrect version</h3></font>
"""

_MISSING_MPCONFIG_ERR = """
<h3><font color='orange'>Warning: no <i>msticpyconfig.yaml</i> found</h3></font>
Some functionality (such as Threat Intel lookups) will not function without
valid configuration settings.
Please go to the <a href="#Configuration">Configuration section</a>
follow the instructions there.
"""

_PANDAS_REQ_VERSION = (0, 25, 0)


def _get_verbose_setting() -> Callable[[Optional[bool]], bool]:
    """Closure for holding trace setting."""
    _verbose_enabled = False

    def _verbose(verbose: Optional[bool] = None) -> bool:
        nonlocal _verbose_enabled
        if verbose is not None:
            _verbose_enabled = verbose
        return _verbose_enabled

    return _verbose


_VERBOSE = _get_verbose_setting()

_NB_IMPORTS = [
    dict(pkg="pandas", alias="pd"),
    dict(pkg="IPython", tgt="get_ipython"),
    dict(pkg="IPython.display", tgt="display"),
    dict(pkg="IPython.display", tgt="HTML"),
    dict(pkg="IPython.display", tgt="Markdown"),
    dict(pkg="ipywidgets", alias="widgets"),
    dict(pkg="pathlib", tgt="Path"),
    dict(pkg="matplotlib.pyplot", alias="plt"),
    dict(pkg="matplotlib", tgt="MatplotlibDeprecationWarning"),
    dict(pkg="seaborn", alias="sns"),
    dict(pkg="numpy", alias="np"),
]
_MP_IMPORTS = [
    dict(pkg="msticpy.data", tgt="QueryProvider"),
    dict(pkg="msticpy.nbtools.foliummap", tgt="FoliumMap"),
    dict(pkg="msticpy.common.utility", tgt="md"),
    dict(pkg="msticpy.common.utility", tgt="md_warn"),
    dict(pkg="msticpy.common.wsconfig", tgt="WorkspaceConfig"),
]
_MP_IMPORT_ALL = [
    dict(module_name="msticpy.nbtools"),
    dict(module_name="msticpy.sectools"),
]

_CONF_URI = (
    "https://msticpy.readthedocs.io/en/latest/getting_started/msticpyconfig.html"
)

_AZNB_GUIDE = (
    "Please see the <i>Getting Started Guide for Azure Sentinel "
    + "ML Notebooks</i> notebook"
)


def init_notebook(
    namespace: Dict[str, Any],
    def_imports: str = "all",
    additional_packages: List[str] = None,
    user_install: bool = False,
    extra_imports: List[str] = None,
    friendly_exceptions: Optional[bool] = None,
    verbose: bool = False,
) -> bool:
    """
    Initialize the notebook environment.

    Parameters
    ----------
    namespace : Dict[str, Any]
        Namespace (usually globals()) into which imports
        are to be populated.
    def_imports : str, optional
        Import default packages. By default "all".
        Possible values are:
        - "all" - import all packages
        - "nb" - import common notebook packages
        - "msticpy" - import msticpy packages
        - "none" (or any other value) don't load any default packages.
    additional_packages : List[str], optional
        Additional packages to be pip installed,
        by default None.
        Packages are specified by name only or version
        specification (e.g. "pandas>=0.25")
    user_install : bool, optional
        Install packages in the "user" rather than system site-packages.
        Use this option if you cannot or do not want to update the system
        packages.
        You should usually avoid using this option with standard Conda environments.
    extra_imports : List[str], optional
        Additional import definitions, by default None.
        Imports are specified as up to 3 comma-delimited values
        in a string:
        "{source_pkg}, [{import_tgt}], [{alias}]"
        `source_pkg` is mandatory - equivalent to a simple "import xyz"
        statement.
        `{import_tgt}` specifies an object to import from the package
        equivalent to "from source_pkg import import_tgt"
        `alias` allows renaming of the imported object - equivalent to
        the "as alias" part of the import statement.
        If you want to provide just `source_pkg` and `alias` include
        an additional placeholder comma: e.g. "pandas, , pd"
    friendly_exceptions : Optional[bool]
        Setting this to True causes msticpy to hook the notebook
        exception hander. Any exceptions derived from MsticpyUserException
        are displayed but do not produce a stack trace, etc.
        Defaults to system/user settings if no value is supplied.
    verbose : bool, optional
        Display more verbose status, by default False

    Returns
    -------
    bool
        True if successful

    Raises
    ------
    MsticpyException
        If extra_imports data format is incorrect.
        If package with required version check has no version
        information.

    """
    _VERBOSE(verbose)

    print("Processing imports....")
    imp_ok = _global_imports(
        namespace, additional_packages, user_install, extra_imports, def_imports
    )

    print("Checking configuration....")
    conf_ok, _ = _check_config()

    print("Setting notebook options....")
    _set_nb_options(namespace)

    if friendly_exceptions is None:
        friendly_exceptions = get_config("msticpy.FriendlyExceptions")
    if friendly_exceptions:
        if verbose:
            print("Friendly exceptions enabled.")
        InteractiveShell.showtraceback = _hook_ipython_exceptions(
            InteractiveShell.showtraceback
        )
    if not imp_ok or not conf_ok:
        md("<font color='red'><h3>Notebook setup did not complete successfully.</h3>")
        if not imp_ok:
            md("One or more libraries did not import successfully.")
            md(_AZNB_GUIDE)
        if not conf_ok:
            md("One or more configuration items were missing or set incorrectly.")
            md(
                _AZNB_GUIDE
                + f" and the <a href='{_CONF_URI}'>msticpy configuration guide</a>."
            )
        md("This notebook may still run but with reduced functionality.")
        return False
    display(HTML("<h3>Notebook setup complete</h3>"))
    return True


def list_default_imports():
    """List the default imports for `init_notebook`."""
    for imp_group in (_NB_IMPORTS, _MP_IMPORTS):
        for imp_item in imp_group:
            if "tgt" in imp_item:
                import_line = f"from {imp_item['pkg']} import {imp_item['tgt']}"
            else:
                import_line = f"import {imp_item['pkg']}"
            if "alias" in imp_item:
                import_line += f" as {imp_item['alias']}"
            print(import_line)
    for imp_item in _MP_IMPORT_ALL:
        print(f"from {imp_item['module_name']} import *")


def _global_imports(  # noqa: MC0001
    namespace: Dict[str, Any],
    additional_packages: List[str] = None,
    user_install: bool = False,
    extra_imports: List[str] = None,
    def_imports: str = "all",
):
    try:
        if def_imports.casefold() in ["all", "nb"]:
            for imp_pkg in _NB_IMPORTS:
                _imp_from_package(nm_spc=namespace, **imp_pkg)

            _check_and_reload_pkg(namespace, pd, _PANDAS_REQ_VERSION, "pd")

        if def_imports.casefold() in ["all", "msticpy"]:
            for imp_pkg in _MP_IMPORTS:
                _imp_from_package(nm_spc=namespace, **imp_pkg)
            for imp_pkg in _MP_IMPORT_ALL:
                _imp_module_all(nm_spc=namespace, **imp_pkg)

        if additional_packages:
            pkg_success = check_and_install_missing_packages(
                additional_packages, user=user_install
            )
            if not pkg_success:
                print("One or more packages failed to install.")
                print(
                    "Please re-run init_notebook() with the parameter user_install=True."
                )
        if extra_imports:
            _import_extras(nm_spc=namespace, extra_imports=extra_imports)
        return True
    except ImportError as imp_err:
        display(HTML(_IMPORT_ERR_MSSG.format(err=imp_err)))
        return False


def _check_config() -> Tuple[bool, Optional[Tuple[List[str], List[str]]]]:
    config_ok = True
    err_warn = None
    mp_path = os.environ.get("MSTICPYCONFIG", "./msticpyconfig.yaml")
    if not Path(mp_path).exists():
        display(HTML(_MISSING_MPCONFIG_ERR))
    else:
        try:
            err_warn = validate_config(config_file=mp_path)
            if err_warn and err_warn[0]:
                config_ok = False
        # pylint: disable=broad-except
        except Exception as err:
            config_ok = False
            print(f"Exception while checking configuration:\n{err}")
        # pylint: enable=broad-except
    ws_config = WorkspaceConfig()
    if not ws_config.config_loaded:
        print("No valid configuration for Azure Sentinel found.")
        config_ok = False
    return config_ok, err_warn


def _set_nb_options(namespace):
    namespace["WIDGET_DEFAULTS"] = {
        "layout": widgets.Layout(width="95%"),
        "style": {"description_width": "initial"},
    }

    # Some of our dependencies (networkx) still use deprecated Matplotlib
    # APIs - we can't do anything about it, so suppress them from view
    warnings.simplefilter("ignore", category=MatplotlibDeprecationWarning)
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    sns.set()
    pd.set_option("display.max_rows", 100)
    pd.set_option("display.max_columns", 50)
    pd.set_option("display.max_colwidth", 100)
    os.environ["KQLMAGIC_LOAD_MODE"] = "silent"


def _import_extras(nm_spc: Dict[str, Any], extra_imports: List[str]):
    for imp_spec in extra_imports:
        params: List[Optional[str]] = [None, None, None]
        for idx, param in enumerate(imp_spec.split(",")):
            params[idx] = param.strip() or None

        if params[0] is None:
            raise MsticpyException(
                f"First parameter in extra_imports is mandatory: {imp_spec}"
            )
        _imp_from_package(nm_spc=nm_spc, pkg=params[0], tgt=params[1], alias=params[2])


def _imp_module(nm_spc: Dict[str, Any], module_name: str, alias: str = None):
    """Import named module and assign to global alias."""
    try:
        mod = importlib.import_module(module_name)
    except ImportError:
        display(HTML(_IMPORT_MODULE_MSSG.format(module=module_name)))
        raise
    if alias:
        nm_spc[alias] = mod
    else:
        nm_spc[module_name] = mod
    if _VERBOSE():  # type: ignore
        print(f"{module_name} imported (alias={alias})")
    return mod


def _imp_module_all(nm_spc: Dict[str, Any], module_name):
    """Import all from named module add to globals."""
    try:
        imported_mod = importlib.import_module(module_name)
    except ImportError:
        display(HTML(_IMPORT_MODULE_MSSG.format(module=module_name)))
        raise
    for item in dir(imported_mod):
        if item.startswith("_"):
            continue
        nm_spc[item] = getattr(imported_mod, item)
    if _VERBOSE():  # type: ignore
        print(f"All items imported from {module_name}")


def _imp_from_package(
    nm_spc: Dict[str, Any], pkg: str, tgt: str = None, alias: str = None
):
    """Import object or submodule from `pkg`."""
    if not tgt:
        return _imp_module(nm_spc=nm_spc, module_name=pkg, alias=alias)
    try:
        # target could be a module
        obj = importlib.import_module(f".{tgt}", pkg)
    except (ImportError, ModuleNotFoundError):
        # if not, it must be an attribute (class, func, etc.)
        try:
            mod = importlib.import_module(pkg)
        except ImportError:
            display(HTML(_IMPORT_MODULE_MSSG.format(module=pkg)))
            raise
        obj = getattr(mod, tgt)
    if alias:
        nm_spc[alias] = obj
    else:
        nm_spc[tgt] = obj
    if _VERBOSE():  # type: ignore
        print(f"{tgt} imported from {pkg} (alias={alias})")
    return obj


def _check_and_reload_pkg(
    nm_spc: Dict[str, Any], pkg: Any, req_version: Tuple[int, ...], alias: str = None
):
    """Check package version matches required version and reload."""
    warn_mssg = []
    pkg_name = pkg.__name__
    if not hasattr(pkg, "__version__"):
        raise MsticpyException(f"Package {pkg_name} has no version data.")
    pkg_version = tuple([int(v) for v in pkg.__version__.split(".")])
    if pkg_version < req_version:
        display(HTML(_MISSING_PKG_WARN.format(package=pkg_name)))
        if not unit_testing():
            resp = input("Install the package now? (y/n)")  # nosec
        else:
            resp = "y"
        if resp.casefold().startswith("y"):
            warn_mssg.append(f"{pkg_name} was installed or upgraded.")
            pip_ver = ".".join([str(elem) for elem in req_version])
            pkg_spec = f"{pkg_name}>={pip_ver}"
            check_and_install_missing_packages(required_packages=[pkg_spec], user=True)

            if pkg_name in sys.modules:
                importlib.reload(pkg)
            else:
                _imp_module(nm_spc, pkg_name, alias=alias)
    if _VERBOSE():  # type: ignore
        print(f"{pkg_name} imported version {pkg.__version__}")
    return warn_mssg


def _hook_ipython_exceptions(func):
    """Hooks the `func` and bypasses it if exception is MsticpyUserException."""

    @wraps(func)
    def showtraceback(*args, **kwargs):
        """Replace IPython showtraceback."""
        # extract exception type, value and traceback
        e_type, _, _ = sys.exc_info()
        if e_type is not None and issubclass(e_type, MsticpyUserError):
            return None
        # otherwise run the original hook
        value = func(*args, **kwargs)
        return value

    return showtraceback



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""."""
import json
import zipfile
from collections import defaultdict
from datetime import datetime
from json import JSONDecodeError
from pathlib import Path
from typing import Any, Dict, Generator, Iterable, List, Optional, Set, Tuple, Union
from zipfile import ZipFile

import attr
import pandas as pd
import requests
import yaml
from tqdm.auto import tqdm

from ..._version import VERSION
from ...common.exceptions import MsticpyNotConnectedError
from ..query_source import QuerySource
from .driver_base import DriverBase

__version__ = VERSION
__author__ = "Ian Hellen"


_MORDOR_TREE_URI = (
    "https://api.github.com/repos/OTRF/mordor/git/trees/master?recursive=1"
)

_MTR_TAC_CAT_URI = "https://attack.mitre.org/tactics/{cat}/"
_MTR_TECH_CAT_URI = "https://attack.mitre.org/techniques/{cat}/"

MITRE_TECHNIQUES: pd.DataFrame = None
MITRE_TACTICS: pd.DataFrame = None


# pylint: disable=too-many-instance-attributes
class MordorDriver(DriverBase):
    """Mordor data driver."""

    def __init__(self, **kwargs):
        """Initialize the Morder driver."""
        super().__init__(**kwargs)
        self.use_query_paths = False
        self.has_driver_queries = True
        self.mitre_techniques: pd.DataFrame
        self.mitre_tactics: pd.DataFrame
        self.mordor_data: Dict[str, MordorEntry]
        self.mdr_idx_tech: Dict[str, Set[str]]
        self.mdr_idx_tact: Dict[str, Set[str]]
        self._driver_queries: List[Dict[str, Any]] = []

        self.use_cached = kwargs.pop("used_cached", True)
        self.save_folder = kwargs.pop("save_folder", ".")
        self.silent = kwargs.pop("silent", False)

        self._loaded = True

    # pylint: disable=global-statement

    def connect(self, connection_str: Optional[str] = None, **kwargs):
        """
        Connect to data source.

        Parameters
        ----------
        connection_str : Optional[str]
            Connect to a data source

        """
        global MITRE_TECHNIQUES, MITRE_TACTICS
        print("Retrieving Mitre data...")

        if MITRE_TECHNIQUES is None:
            MITRE_TECHNIQUES = _get_mitre_categories(_MTR_TECH_CAT_URI)
        self.mitre_techniques = MITRE_TECHNIQUES
        if MITRE_TACTICS is None:
            MITRE_TACTICS = _get_mitre_categories(_MTR_TAC_CAT_URI)
        self.mitre_tactics = MITRE_TACTICS

        print("Retrieving Mordor data...")
        self.mordor_data = _GET_MORDOR_METADATA()
        self.mdr_idx_tech, self.mdr_idx_tact = _build_mdr_indexes(self.mordor_data)

        self._connected = True
        self.public_attribs = {
            "mitre_techniques": self.mitre_techniques,
            "mitre_tactics": self.mitre_tactics,
            "driver_queries": self.driver_queries,
            "search_queries": self.search_queries,
        }

    # pylint: enable=global-statement

    def query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Other Parameters
        ----------------
        kwargs :
            Are passed to the underlying provider query method,
            if supported.

        Returns
        -------
        Union[pd.DataFrame, Any]
            A DataFrame (if successfull) or
            the underlying provider result if an error.

        """
        del query_source
        if not self._connected:
            raise self._create_not_connected_err()
        use_cached = kwargs.pop("used_cached", self.use_cached)
        save_folder = kwargs.pop("save_folder", self.save_folder)
        silent = kwargs.pop("silent", self.silent)
        result_df = download_mdr_file(
            file_uri=query,
            use_cached=use_cached,
            save_folder=save_folder,
            silent=silent,
        )
        if not isinstance(result_df, pd.DataFrame) or result_df.empty:
            return "Could not convert result to a DataFrame."
        return result_df

    def query_with_results(self, query: str, **kwargs) -> Tuple[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame plus native results.

        Parameters
        ----------
        query : str
            The query to execute

        Returns
        -------
        Tuple[pd.DataFrame,Any]
            A DataFrame and native results.

        """
        result = self.query(query, **kwargs)
        if isinstance(result, pd.DataFrame):
            return result, "OK"
        return pd.DataFrame, result

    @property
    def driver_queries(self) -> Iterable[Dict[str, Any]]:
        """
        Return generator of Mordor query definitions.

        Yields
        ------
        Iterable[Dict[str, Any]]
            Iterable of Dictionaries containing query definitions.

        """
        if not self._connected:
            raise self._create_not_connected_err()
        if not self._driver_queries:
            self._driver_queries = list(self._get_driver_queries())
        return self._driver_queries

    def _get_driver_queries(self):
        """Generate iterable of Mordor queries."""
        for mdr_item in self.mordor_data.values():
            for file_path in mdr_item.get_file_paths():
                mitre_data = mdr_item.get_attacks()
                techniques = ", ".join(
                    [f"{att.technique}: {att.technique_name}" for att in mitre_data]
                )
                tactics = ", ".join(
                    [
                        f"{tac[0]}: {tac[1]}"
                        for att in mitre_data
                        for tac in att.tactics_full
                    ]
                )
                doc_string: List[str] = [
                    f"{mdr_item.title}",
                    "",
                    "Notes",
                    "-----",
                    f"Mordor ID: {mdr_item.id}",
                    mdr_item.description or "",
                    "",
                    f"Mitre Techniques: {techniques}",
                    f"Mitre Tactics: {tactics}",
                ]
                q_container, _, full_name = file_path["qry_path"].partition(".")
                short_name = file_path["qry_path"].split(".")[-1]
                yield {
                    "name": full_name,
                    "description": "\n".join(doc_string),
                    "query_name": short_name,
                    "query": file_path["file_path"],
                    "query_container": q_container,
                    "metadata": {},
                }

    def search_queries(self, search: str) -> Iterable[str]:
        """
        Search queries for matching attributes.

        Parameters
        ----------
        search : str
            Search string. Substrings separated by commas will
            be treated as OR terms - e.g. "a, b" == "a" or "b".
            Substrings separated by "+" will be treated as AND
            terms - e.g. "a + b" == "a" and "b"

        Returns
        -------
        Iterable[str]
            Iterable of matching query names.


        """
        if not self._connected:
            raise self._create_not_connected_err()
        matches = []
        for mdr_id in search_mdr_data(self.mordor_data, terms=search):
            for file_path in self.mordor_data[mdr_id].get_file_paths():
                matches.append(
                    f"{file_path['qry_path']} ({self.mordor_data[mdr_id].title})"
                )
        return matches

    @staticmethod
    def _create_not_connected_err():
        return MsticpyNotConnectedError(
            "Please run the connect() method before running this method.",
            title="not connected to Mordor.",
            help_uri="https://msticpy.readthedocs.io/en/latest/DataProviders.html",
        )


# pylint: enable=too-many-instance-attributes


class MitreAttack:
    """MitreAttack container for techniques and tactics."""

    MTR_TECH_URI = "https://attack.mitre.org/techniques/{technique_id}/"
    MTR_TAC_URI = "https://attack.mitre.org/tactics/{tactic_id}/"

    def __init__(
        self,
        attack: Dict[str, Any] = None,
        technique: str = None,
        sub_technique: str = None,
        tactics: List[str] = None,
    ):
        """
        Create instance of MitreAttack.

        Parameters
        ----------
        attack : Dict[str, Any], optional
            attack data as dictionary, by default None
        technique : str, optional
            technique ID, by default None
        sub_technique : str, optional
            sub-technique ID, by default None
        tactics : List[str], optional
            List of associated tactics, by default None

        """
        if attack is None and (technique is None and tactics is None):
            raise TypeError(
                "Either 'attack' or 'technique' and 'tactics' must be specified."
            )
        self.technique = attack.get("technique") if attack else technique
        self.sub_technique = attack.get("sub-technique") if attack else sub_technique
        self.tactics = attack.get("tactics") if attack else tactics  # type: ignore

        self._technique_name = None
        self._technique_desc = None
        self._technique_uri = None
        self._tactics_full: List[Tuple[str, str, str, str]] = []

    def __repr__(self) -> str:
        """
        Return repr of MitreAttack object.

        Returns
        -------
        str
            The repr of the object.

        """
        return "".join(
            [
                f"MitreAttack(technique={self.technique}), ",
                f"sub_technique={self.sub_technique}, ",
                f"tactics={repr(self.tactics)}",
            ]
        )

    @property
    def technique_name(self) -> Optional[str]:
        """
        Return Mitre Technique full name.

        Returns
        -------
        Optional[str]
            Name of the Mitre technique

        """
        if not self._technique_name and self.technique in MITRE_TECHNIQUES.index:
            self._technique_name = MITRE_TECHNIQUES.loc[self.technique].Name
        return self._technique_name

    @property
    def technique_desc(self) -> Optional[str]:
        """
        Return Mitre technique description.

        Returns
        -------
        Optional[str]
            Technique description

        """
        if not self._technique_desc and self.technique in MITRE_TECHNIQUES.index:
            self._technique_desc = MITRE_TECHNIQUES.loc[self.technique].Description
        return self._technique_desc

    @property
    def technique_uri(self) -> str:
        """
        Return Mitre Technique URI.

        Returns
        -------
        Optional[str]
            URI of the Mitre technique

        """
        return self.MTR_TECH_URI.format(technique_id=self.technique)

    @property
    def tactics_full(self) -> List[Tuple[str, str, str, str]]:
        """
        Return full listing of Mitre tactics.

        Returns
        -------
        List[Tuple[str, str, str, str]]
            List of tuples of:
            (ID, Name, Description, URI)

        """
        if not self._tactics_full and self.tactics:
            for tactic in self.tactics:
                tactic_name = tactic_desc = "unknown"
                if tactic in MITRE_TACTICS.index:
                    tactic_name = MITRE_TACTICS.loc[tactic].Name
                    tactic_desc = MITRE_TACTICS.loc[tactic].Description
                tactic_uri = self.MTR_TAC_URI.format(tactic_id=tactic)
                self._tactics_full.append(
                    (tactic, tactic_name, tactic_desc, tactic_uri)
                )
        return self._tactics_full


def _to_datetime(date_val) -> datetime:
    """
    Return datetime from parsed date string.

    Parameters
    ----------
    date_val : datetime
        The datetime or datetime string.

    Returns
    -------
    datetime
        Parse datetime.

    """
    if isinstance(date_val, datetime):
        return date_val
    try:
        return pd.to_datetime(date_val)
    except TypeError:
        return datetime.min


DS_PREFIX = "https://raw.githubusercontent.com/OTRF/mordor/master/datasets/"


# pylint: disable=not-an-iterable, no-member


@attr.s(auto_attribs=True)
class MordorEntry:
    """Mordor data set metadata."""

    title: str
    id: str
    author: str
    creation_date: datetime = attr.ib(converter=_to_datetime)
    modification_date: datetime = attr.ib(converter=_to_datetime)
    platform: Optional[str] = None
    description: Optional[str] = None
    tags: List[str] = attr.Factory(list)
    files: List[Dict[str, Any]] = attr.Factory(list)
    datasets: List[Dict[str, Any]] = attr.Factory(list)
    attack_mappings: List[Dict[str, Any]] = attr.Factory(list)
    notebooks: List[Dict[str, str]] = attr.Factory(list)
    simulation: Dict[str, Any] = attr.Factory(dict)
    references: List[Any] = attr.Factory(list)
    _rel_file_paths: List[Dict[str, Any]] = attr.Factory(list)

    def get_notebooks(self) -> List[Tuple[str, str, str]]:
        """
        Return the list of notebooks for the dataset.

        Returns
        -------
        List[Tuple[str, str, str]]
            Tuples of (name, project, link)

        """
        return [
            (nbk.get("name", ""), nbk.get("project", ""), nbk.get("link", ""))
            for nbk in self.notebooks
        ]

    def get_attacks(self) -> List[MitreAttack]:
        """
        Return list of Mitre attack classifications.

        Returns
        -------
        List[MitreAttack]
            List of MitreAttack definitions.

        """
        return [MitreAttack(attack=attack) for attack in self.attack_mappings]

    def get_file_paths(self) -> List[Dict[str, str]]:
        """
        Return list of data file links.

        Returns
        -------
        List[Dict[str, str]]
            list of dictionaries describing files.
            Each entry has key/values for:
            - file_type
            - file_path
            - relative_path
            - qry_path

        """
        if not self._rel_file_paths:
            for file in self.files:
                f_path = file.get("link")
                if not f_path:
                    continue
                f_rel_path = f_path.replace(DS_PREFIX, "")
                query_path = ".".join(Path(f_rel_path).parts).replace(
                    Path(f_rel_path).suffix, ""
                )
                self._rel_file_paths.append(
                    {
                        "file_type": file.get("type"),
                        "file_path": f_path,
                        "relative_path": f_rel_path,
                        "qry_path": query_path,
                    }
                )
        return self._rel_file_paths


# pylint: disable=not-an-iterable, no-member


def get_mdr_data_paths(item_type="metadata") -> Generator[str, None, None]:
    """
    Generate Mordor data sets from GitHub repo.

    Parameters
    ----------
    item_type : str, optional
        The type of item required, by default "metadata"
        Other values are "large", "small.

    Yields
    ------
    str
        Iterable of paths

    """
    md_tree = _GET_MORDOR_TREE(_MORDOR_TREE_URI)
    prefix = f"datasets/{item_type}"
    yield from (
        t_item["path"]
        for t_item in md_tree.get("tree")
        if t_item["type"] == "blob" and t_item["path"].startswith(prefix)
    )


def _get_mdr_github_tree():
    """Closure to wrap fetching Mordor tree from GitHub."""
    mordor_tree = None

    def _get_mdr_tree(uri):
        nonlocal mordor_tree
        if mordor_tree is None:
            resp = requests.get(uri)
            mordor_tree = resp.json()
        return mordor_tree

    return _get_mdr_tree


# Create closure
_GET_MORDOR_TREE = _get_mdr_github_tree()


def _get_mdr_file(gh_file):
    """Fetch a file from Mordor repo."""
    file_blob_uri = f"https://raw.githubusercontent.com/OTRF/mordor/master/{gh_file}"
    file_resp = requests.get(file_blob_uri)
    return file_resp.content


def _create_mdr_metadata_cache():
    md_metadata: Dict[str, MordorEntry] = {}

    def _get_mdr_metadata():
        nonlocal md_metadata
        if not md_metadata:
            md_metadata = _fetch_mdr_metadata()
        return md_metadata

    return _get_mdr_metadata


# Create closure
_GET_MORDOR_METADATA = _create_mdr_metadata_cache()


# pylint: disable=global-statement
def _fetch_mdr_metadata() -> Dict[str, MordorEntry]:
    """
    Return full metadata for Mordor datasets.

    Returns
    -------
    Dict[str, MordorEntry]:
        Mordor data set metadata keyed by MordorID

    """
    global MITRE_TECHNIQUES, MITRE_TACTICS

    if MITRE_TECHNIQUES is None:
        MITRE_TECHNIQUES = _get_mitre_categories(_MTR_TECH_CAT_URI)
    if MITRE_TACTICS is None:
        MITRE_TACTICS = _get_mitre_categories(_MTR_TAC_CAT_URI)

    md_metadata: Dict[str, MordorEntry] = {}
    mdr_md_paths = list(get_mdr_data_paths("metadata"))
    for y_file in tqdm(mdr_md_paths, unit=" files", desc="Downloading Mordor metadata"):
        gh_file_content = _get_mdr_file(y_file)
        yaml_doc = yaml.safe_load(gh_file_content)
        doc_id = yaml_doc.get("id")
        md_metadata[doc_id] = MordorEntry(**yaml_doc)
    return md_metadata


# pylint: enable=global-statement


def _build_mdr_indexes(
    mdr_metadata: Dict[str, MordorEntry]
) -> Tuple[Dict[str, Set[str]], Dict[str, Set[str]]]:
    """
    Return dictionaries mapping Mitre items to Mordor datasets.

    Parameters
    ----------
    mdr_metadata : Dict[str, MordorEntry]
        Dictionary of mordor dataset metadata.

    Returns
    -------
    Tuple[Dict[str, Set[str]], Dict[str, Set[str]]]
        Mordor dataset IDs keyed by Mitre Technique and Mitre Tactic.

    """
    md_idx_techniques = defaultdict(set)
    md_idx_tactics = defaultdict(set)

    for md_id, md_file in mdr_metadata.items():
        for attack in md_file.get_attacks():
            md_idx_techniques[attack.technique].add(md_id)
            if not attack.tactics:
                continue
            for tactic in attack.tactics:
                md_idx_tactics[tactic].add(md_id)

    return md_idx_techniques, md_idx_tactics  # type: ignore


def download_mdr_file(
    file_uri: str, use_cached: bool = True, save_folder: str = ".", silent: bool = False
) -> pd.DataFrame:
    """
    Download data file from Mordor.

    Parameters
    ----------
    file_uri : str
        The URI of the file to download.
    use_cached : bool, optional
        Try to use locally saved file first, by default True
    save_folder : str, optional
        Path to output folder, by default "."
    silent : bool
        If True, suppress feedback. By default, False.

    Returns
    -------
    pd.DataFrame
        DataFrame of Dataset

    """
    if not silent:
        print(file_uri)
    if not file_uri.lower().endswith("zip"):
        raise TypeError(f"File type not supported {file_uri}")
    if not Path(save_folder).is_dir():
        Path(save_folder).mkdir(parents=True, exist_ok=True)
    save_path = "-".join(Path(file_uri.replace(DS_PREFIX, "")).parts)
    save_file = Path(save_folder).joinpath(save_path)
    if not use_cached or not save_file.is_file():
        # streamed download
        resp = requests.get(file_uri, stream=True)
        with open(str(save_file), "wb") as fdesc:
            for chunk in resp.iter_content(chunk_size=1024):
                fdesc.write(chunk)

    zip_file = zipfile.ZipFile(str(save_file))
    file_names = zip_file.namelist()
    d_frames = {}
    for file_name in file_names:
        d_frames[file_name] = _extract_zip_file_to_df(
            zip_file, file_name, use_cached, save_folder, silent
        )

    return pd.concat(d_frames.values())


def _extract_zip_file_to_df(  # noqa: MC0001
    zip_file: ZipFile,
    file_name: str,
    use_cached: bool = True,
    save_folder: str = ".",
    silent: bool = False,
) -> pd.DataFrame:
    """
    Extract from zip and parse json file to DataFrame.

    Parameters
    ----------
    zip_file : ZipFile
        ZipFile object containing the file
    file_name : str
        File name to extract
    use_cached : bool, optional
        Try to use locally saved file first, by default True
    save_folder : str, optional
        Path to output folder, by default "."
    silent : bool
        If False, suppress feedback. By default, True.

    Returns
    -------
    pd.DataFrame
        Extracted DataFrame

    """
    if not silent:
        print("Extracting", file_name)

    file_path = Path(save_folder).joinpath(file_name)
    if not use_cached or not file_path.is_file():
        zip_file.extract(file_name, path=save_folder)

    out_df = pd.DataFrame()
    if file_path.suffix.lower() == ".json":
        out_df = pd.read_json(file_path, lines=True)
    if file_path.suffix.lower() == ".csv":
        out_df = pd.read_csv(file_path)
    if file_path.suffix.lower() not in (".json", ".csv"):
        print(f"Cannot process files of type {file_path.suffix.lower()}")
    if not use_cached:
        Path(file_name).unlink()
    return out_df


def _json_to_df(file_path, silent):
    errs = []
    with open(str(file_path), "r") as j_file:
        j_text = j_file.read()
    df_list = []
    if silent:
        line_gen = enumerate(j_text.split("\n"))
    else:
        line_gen = tqdm(enumerate(j_text.split("\n")), "lines")
    for line_num, line in line_gen:
        if not line:
            continue
        try:
            df_list.append(json.loads(line))
        except JSONDecodeError:
            errs.append(f"Could not parse #{line_num}: '{line}'")
    out_df = pd.DataFrame(df_list)
    if errs:
        print(f"{len(errs)} errors detected", errs)
    return out_df


def search_mdr_data(
    mdr_data: Dict[str, MordorEntry], terms: str = None, subset: Iterable[str] = None
) -> Set[str]:
    """
    Return IDs for items matching terms.

    Parameters
    ----------
    mdr_data : Dict[str, MordorEntry]
        Mordor dataset
    terms : str, optional
        Search terms, by default None
        (comma-separated values are treated as OR terms
        plus-separated values are treated as AND terms)
    subset : Iterable[str], optional
        A subset of IDs over which to search, by default None

    Returns
    -------
    Set[str]
        The set of matching IDs.

    """
    if terms is None:
        return set(subset or mdr_data.keys())
    logic = "OR"
    if "," in terms:
        search_terms = terms.split(",")
    elif "+" in terms:
        search_terms = terms.split("+")
        logic = "AND"
    else:
        search_terms = [terms]
    results: Set[str] = set()
    for search_idx, term in enumerate(search_terms):
        item_results = set()
        for md_id, item in mdr_data.items():
            if subset is not None and md_id not in subset:
                continue
            if term.strip() in str(item):
                item_results.add(md_id)
        if logic == "OR":
            results = results | item_results
        else:
            # Don't AND if search_idx == 0 (and-ing against empty results)
            results = results & item_results if search_idx else item_results
    return results


def _get_mitre_categories(uri_template: str) -> pd.DataFrame:
    """
    Download and return Mitre techniques and tactics.

    Parameters
    ----------
    uri_template : str
        URI to fetch MITRE category from.

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame]
        Tuple of techniques, tactics with IDs, names,
        descriptions.

    """
    categories = {"pre": "PreAttack", "enterprise": "Enterprise", "mobile": "Mobile"}

    mitre_data = pd.concat(
        [
            pd.read_html(uri_template.format(cat=cat))[0].assign(MitreGroup=cat_title)
            for cat, cat_title in categories.items()
        ]
    ).set_index("ID")
    if "ID.1" in mitre_data.columns:
        mitre_data.drop(columns=["ID.1"])

    return mitre_data



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for common display functions."""
from typing import Any, Mapping, Union, Tuple, List

import matplotlib.pyplot as plt
import networkx as nx
import pandas as pd

import IPython
from IPython.core.display import HTML, display
from IPython.display import Javascript

from .._version import VERSION
from .security_alert import SecurityAlert

# pylint: disable=unused-import
from .timeline import display_timeline, display_timeline_values  # noqa
from .process_tree import build_and_show_process_tree, plot_process_tree  # noqa

# pylint: enable=unused-import
from ..common.utility import export

__version__ = VERSION
__author__ = "Ian Hellen"


@export
def display_alert(
    alert: Union[Mapping[str, Any], SecurityAlert], show_entities: bool = False
):
    """
    Display a Security Alert.

    Parameters
    ----------
    alert : Union[Mapping[str, Any], SecurityAlert]
        The alert to display as Mapping (e.g. pd.Series)
        or SecurityAlert
    show_entities : bool, optional
        Whether to display entities (the default is False)

    """
    output = format_alert(alert, show_entities)
    if not isinstance(output, tuple):
        output = [output]
    for disp_obj in output:
        display(disp_obj)


@export
def format_alert(
    alert: Union[Mapping[str, Any], SecurityAlert], show_entities: bool = False
) -> Union[IPython.display.HTML, Tuple[IPython.display.HTML, pd.DataFrame]]:
    """
    Get IPython displayable Security Alert.

    Parameters
    ----------
    alert : Union[Mapping[str, Any], SecurityAlert]
        The alert to display as Mapping (e.g. pd.Series)
        or SecurityAlert
    show_entities : bool, optional
        Whether to display entities (the default is False)

    Returns
    -------
    Union[IPython.display.HTML, Tuple[IPython.display.HTML, pd.DataFrame]]
        Single or tuple of displayable IPython objects

    Raises
    ------
    ValueError
        If the alert object is in an unknown format

    """
    if isinstance(alert, SecurityAlert):
        return HTML(alert.to_html(show_entities=show_entities))

    # Display subset of raw properties
    if isinstance(alert, pd.Series):
        entity = alert["CompromisedEntity"] if "CompromisedEntity" in alert else ""
        title = f"""
            <h3>Series Alert: '{alert["AlertDisplayName"]}'</h3>
            <b>Alert_time:</b> {alert["StartTimeUtc"]},&nbsp;
            <b>Compr_entity:</b> {entity},&nbsp;
            <b>Alert_id:</b> {alert["SystemAlertId"]}
            <br/>
            """
        return HTML(title), pd.DataFrame(alert)

    raise ValueError("Unrecognized alert object type " + str(type(alert)))


@export
def display_process_tree(process_tree: pd.DataFrame):
    """
    Display process tree data frame. (Deprecated).

    Parameters
    ----------
    process_tree : pd.DataFrame
        Process tree DataFrame

    The display module expects the columns NodeRole and Level to
    be populated. NoteRole is one of: 'source', 'parent', 'child'
    or 'sibling'. Level indicates the 'hop' distance from the 'source'
    node.

    """
    build_and_show_process_tree(process_tree)


@export
def exec_remaining_cells():
    """Execute all cells below currently selected cell."""
    Javascript("Jupyter.notebook.execute_cells_below()")


@export
# pylint: disable=too-many-arguments
def draw_alert_entity_graph(
    nx_graph: nx.Graph,
    font_size: int = 12,
    height: int = 15,
    width: int = 15,
    margin: float = 0.3,
    scale: int = 1,
):
    """
    Draw networkX graph with matplotlib.

    Parameters
    ----------
    nx_graph : nx.Graph
        The NetworkX graph to draw
    font_size : int, optional
        base font size (the default is 12)
    height : int, optional
        Image height (the default is 15)
    width : int, optional
        Image width (the default is 15)
    margin : float, optional
        Image margin (the default is 0.3)
    scale : int, optional
        Position scale (the default is 1)

    """
    alert_node = [
        n
        for (n, node_type) in nx.get_node_attributes(nx_graph, "node_type").items()
        if node_type == "alert"
    ]
    entity_nodes = [
        n
        for (n, node_type) in nx.get_node_attributes(nx_graph, "node_type").items()
        if node_type == "entity"
    ]

    # now draw them in subsets  using the `nodelist` arg
    plt.rcParams["figure.figsize"] = (width, height)

    plt.margins(x=margin, y=margin)

    pos = nx.kamada_kawai_layout(nx_graph, scale=scale, weight="weight")
    nx.draw_networkx_nodes(
        nx_graph, pos, nodelist=alert_node, node_color="red", alpha=0.5, node_shape="o"
    )
    nx.draw_networkx_nodes(
        nx_graph,
        pos,
        nodelist=entity_nodes,
        node_color="green",
        alpha=0.5,
        node_shape="s",
        node_size=200,
    )
    nlabels = nx.get_node_attributes(nx_graph, "description")
    nx.relabel_nodes(nx_graph, nlabels)
    nx.draw_networkx_labels(nx_graph, pos, nlabels, font_size=font_size)
    nx.draw_networkx_edges(nx_graph, pos)
    elabels = nx.get_edge_attributes(nx_graph, "description")
    nx.draw_networkx_edge_labels(
        nx_graph, pos, edge_labels=elabels, font_size=font_size * 2 / 3, alpha=0.6
    )


# Constants for Windows logon
_WIN_LOGON_TYPE_MAP = {
    0: "Unknown",
    2: "Interactive",
    3: "Network",
    4: "Batch",
    5: "Service",
    7: "Unlock",
    8: "NetworkCleartext",
    9: "NewCredentials",
    10: "RemoteInteractive",
    11: "CachedInteractive",
}
_WINDOWS_SID = {
    "S-1-0-0": "Null SID",
    "S-1-5-18": "LOCAL_SYSTEM",
    "S-1-5-19": "LOCAL_SERVICE",
    "S-1-5-20": "NETWORK_SERVICE",
}
_ADMINISTRATOR_SID = "500"
_GUEST_SID = "501"
_DOM_OR_MACHINE_SID = "S-1-5-21"


@export
def display_logon_data(
    logon_event: pd.DataFrame, alert: SecurityAlert = None, os_family: str = None
):
    """
    Display logon data for one or more events as HTML table.

    Parameters
    ----------
    logon_event : pd.DataFrame
        Dataframe containing one or more logon events
    alert : SecurityAlert, optional
        obtain os_family from the security alert
        (the default is None)
    os_family : str, optional
         explicitly specify os_family (Linux or Windows)
         (the default is None)

    Notes
    -----
    Currently only Windows Logon events.

    """
    display(format_logon(logon_event, alert, os_family))


@export
def format_logon(
    logon_event: Union[pd.DataFrame, pd.Series],
    alert: SecurityAlert = None,
    os_family: str = None,
) -> IPython.display.HTML:
    """
    Return logon data for one or more events as HTML table.

    Parameters
    ----------
    logon_event : Union[pd.DataFrame, pd.Series]
        Dataframe containing one or more logon events
        or Series containing a single logon event.
    alert : SecurityAlert, optional
        obtain os_family from the security alert
        (the default is None)
    os_family : str, optional
         explicitly specify os_family (Linux or Windows)
         (the default is None)

    Returns
    -------
    IPython.display.HTML :
        HTML display object

    """
    if not os_family:
        os_family = alert.os_family if alert else "Windows"

    logon_output = []
    if isinstance(logon_event, pd.DataFrame):
        for _, logon_row in logon_event.iterrows():
            logon_record = _fmt_single_row(logon_row, os_family)
            logon_output.append(
                "<tr class='cell_logon'><td class='cell_logon'>"
                + f"{'<br>'.join(logon_record)}</td></tr>"
            )
    elif isinstance(logon_event, pd.Series):
        logon_record = _fmt_single_row(logon_event, os_family)
        logon_output.append(
            "<tr class='cell_logon'><td class='cell_logon'>"
            + f"{'<br>'.join(logon_record)}</td></tr>"
        )

    t_style = """
        <style>
            .table_logon {border-collapse: collapse; width: 50%;}
            .cell_logon {border: 1px solid #ddd !important;
                text-align: left !important; padding: 15px !important;}
        </style>
        """
    return HTML(f"{t_style}<table class='table_logon'>{''.join(logon_output)}</table>")


def _fmt_single_row(logon_row: pd.Series, os_family: str) -> List[str]:
    """Format a pandas series logon record."""
    logon_record = []
    logon_record.append(f"<b>Account: </b>{logon_row['TargetUserName']}")
    logon_record.append(f"<b>Account Domain: </b>{logon_row['TargetDomainName']}")
    logon_record.append(f"<b>Logon Time: </b>{logon_row['TimeGenerated']}")

    if os_family == "Windows":
        logon_type = logon_row["LogonType"]
        logon_desc_idx = logon_type
        if logon_type not in _WIN_LOGON_TYPE_MAP:
            logon_desc_idx = 0
        logon_record.append(
            f"<b>Logon type: </b>{logon_type}"
            + f"({_WIN_LOGON_TYPE_MAP[logon_desc_idx]})"
        )

    account_id = logon_row.TargetUserSid
    logon_record.append(f"<b>User Id/SID: </b>{account_id}")
    if os_family == "Windows":
        logon_record.extend(_format_sid_info(account_id))
    else:
        logon_record.append(f"<b>Audit user: </b>{logon_row['audit_user']}")

    session_id = logon_row["TargetLogonId"]
    sess_id = f"<b>Session id: </b>'{session_id}'"
    if session_id in ["0x3e7", "-1"]:
        sess_id += "System logon session"
    logon_record.append("")

    domain = logon_row["SubjectDomainName"]
    if not domain:
        subj_account = logon_row.SubjectUserName
    else:
        subj_account = f"{domain}/{logon_row.SubjectUserName}"
    logon_record.append(f"<b>Subject (source) account: </b>{subj_account}")

    logon_record.append(f"<b>Logon process: </b>{logon_row['LogonProcessName']}")
    logon_record.append(
        f"<b>Authentication: </b>{logon_row['AuthenticationPackageName']}"
    )
    logon_record.append(f"<b>Source IpAddress: </b>{logon_row['IpAddress']}")
    logon_record.append(f"<b>Source Host: </b>{logon_row['WorkstationName']}")
    logon_record.append(f"<b>Logon status: </b>{logon_row['Status']}")
    logon_record.append("")
    return logon_record


def _format_sid_info(sid):
    sid_info = []
    if not sid:
        return sid_info
    if sid in _WINDOWS_SID:
        sid_info.append(f"&nbsp;&nbsp;SID {sid} is {_WINDOWS_SID[sid]}")
    elif sid.endswith(_ADMINISTRATOR_SID):
        sid_info.append(f"&nbsp;&nbsp;SID {sid} is administrator")
    elif sid.endswith(_GUEST_SID):
        sid_info.append(f"&nbsp;&nbsp;SID {sid} is guest")
    if sid.startswith(_DOM_OR_MACHINE_SID):
        sid_info.append(f"&nbsp;&nbsp;SID {sid} is local machine or domain account")
    return sid_info



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for SecurityAlert class."""
import html
import re
from collections import Counter
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

import pandas as pd

from .._version import VERSION
from ..common.utility import escape_windows_path, export
from ..data.query_defns import DataEnvironment, DataFamily, QueryParamProvider
from .entityschema import Account, Entity, Host, Process

__version__ = VERSION
__author__ = "Ian Hellen"

_ID_PROPERTIES: List[str] = [
    "AzSubscriptionId",
    "AzResourceId",
    "WorkspaceId",
    "AgentId",
    "TenantId",
    "SourceComputerId",
    "ResourceId",
    "WorkspaceSubscriptionId",
    "WorkspaceResourceGroup",
    "ProviderAlertId",
    "SystemAlertId",
    "ResourceId",
]


# pylint: disable=too-many-public-methods
@export
class SecurityBase(QueryParamProvider):
    """
    Security Base Class for alerts and events.

    Instantiates a security event or alert from a pandas Series
    and provides convenience access methods to retrieve properties.
    """

    def __init__(self, src_row: pd.Series = None):
        """Instantiate a security alert from a pandas Series."""
        self._source_data: pd.Series = src_row
        self._custom_query_params: Dict[str, Any] = {}
        self._entities: List[Entity] = []

        # Extract and cache alert ID properties
        self._ids: Dict[str, str] = dict()
        if self._source_data is not None:
            for id_property in _ID_PROPERTIES:
                if id_property in self._source_data:
                    self._ids[id_property] = self._source_data[id_property]

        self.path_separator = "\\"
        self.os_family = "Windows"

    # Dunder methods
    def __getitem__(self, key):
        """Allow property get using dictionary key syntax."""
        if key in self.__dict__:
            return self.__dict__[key]
        if key in self._source_data:
            return self._source_data[key]
        raise KeyError

    def __setitem__(self, key, value):
        """Allow property set using dictionary key syntax."""
        self.__dict__[key] = value

    def __contains__(self, key):
        """Allow property 'in' test."""
        # get attribute from source dictionary if not defined
        return key in self.__dict__ or key in self._source_data

    def __getattr__(self, name):
        """Return the value of the named property 'name'."""
        if name in self._source_data:
            return self._source_data[name]
        raise AttributeError(f"{name} is not a valid attribute.")

    def __str__(self):
        """Return string representation of object properties."""
        str_props = [f"{prop}: {val}" for prop, val in self._source_data.items()]

        if self.entities:
            str_entities = []
            for ent in self.entities:
                str_entities.append(str(ent).replace("\n", ", "))
            str_props = str_props + str_entities
        return "\n".join(str_props)

    def __repr__(self) -> str:
        """Return repr of item."""
        params = ", ".join([f"{name}={val}" for name, val in self.properties.items()])
        if len(params) > 80:
            params = params[:80] + "..."
        return f"{self.__class__.__name__}({params})"

    def _repr_html_(self) -> str:
        """Display in IPython."""
        return self.to_html()

    # def __getstate__(self):
    #     """Return dictionary of state for serialization/pickling."""
    #     state_dict = {}
    #     state_dict['source_data'] = (self._source_data)
    #     state_dict['Entities'] = self._entities
    #     state_dict['path_separator'] = self.path_separator
    #     state_dict['os_family'] = self.os_family
    #     return state_dict

    # def __setstate__(self, state):
    #     """Set state from dictionary for deserialization/unpickling."""
    #     self._source_data = state['source_data']
    #     self._entities = list(state['Entities'])
    #     self.path_separator = state['path_separator']
    #     self.os_family = state['os_family']

    # Properties
    @property
    def entities(self) -> List[Entity]:
        """
        Return a list of the Alert or Event entities.

        Returns
        -------
        List[Entity]
            List of the Alert or Event entities.

        """
        return self._entities

    @property
    def properties(self) -> Dict[str, Any]:
        """
        Return a dictionary of the Alert or Event properties.

        Returns
        -------
        Dict[str, Any]
            dictionary of the Alert or Event properties.

        """
        return self._source_data.to_dict()

    @property
    def hostname(self) -> str:
        """Return the Hostname (not FQDN) of the host associated with the alert."""
        return self.primary_host.HostName if self.primary_host is not None else None

    @property
    def computer(self) -> Optional[str]:
        """
        Return the Computer name of the host associated with the alert.

        (host FQDN, if available)
        """
        return self.primary_host.computer if self.primary_host is not None else None

    @property
    def ids(self) -> Dict[str, str]:
        """Return a collection of Identity properties for the alert."""
        return self._ids

    @property
    def is_in_workspace(self) -> bool:
        """Return True if the alert has a Log Analytics WorkspaceID."""
        return "WorkspaceId" in self._ids and "AgentId" in self._ids

    @property
    def is_in_log_analytics(self) -> bool:
        """Return True if the alert originates from a Log Analytics Workspace host."""
        return "TenantId" in self._ids

    @property
    def is_in_azure_sub(self) -> bool:
        """Return True if the alert originates from an Azure Security Center host."""
        if (
            "AzSubscriptionId" not in self._ids
            and "AzResourceId" not in self._ids
            and "ResourceId" in self._ids
            and self._ids["ResourceId"]
        ):
            self._ids["AzResourceId"] = self._id["ResourceId"]
            res = self._get_subscription_from_resource(self._id["ResourceId"])
            if res:
                self._ids["AzSubscriptionId"] = res

        return "AzSubscriptionId" in self._ids and "AzResourceId" in self._ids

    @property
    def primary_host(self) -> Optional[Union[Host, Entity]]:
        """
        Return the primary host entity (if any) associated with this object.

        Returns
        -------
        Optional[Host]
            primary host entity (if any)

        """
        hosts = self.get_entities_of_type("host")
        if hosts:
            return hosts[0]
        return None

    @property
    def primary_process(self) -> Optional[Union[Process, Entity]]:
        """
        Return the primary process entity (if any) associated with this object.

        Returns
        -------
        Optional[Process]
            primary process entity (if any)

        """
        procs = self.get_entities_of_type("process")
        if not procs:
            return None
        if len(procs) == 1:
            return procs[0]

        # find the first process that has a parent process property
        procs_with_parent = [p for p in procs if "ParentProcess" in p]
        return procs_with_parent[0] if procs_with_parent else procs[0]

    @property
    def primary_account(self) -> Optional[Union[Process, Entity]]:
        """
        Return the primary account entity (if any) associated with this object.

        Returns
        -------
        Optional[Process]
            primary account entity (if any)

        """
        accts = self.get_entities_of_type("account")
        return accts[0] if accts else None

    @property
    def query_params(self) -> Dict[str, Any]:
        """
        Query parameters derived from alert.

        Returns
        -------
        Dict[str, Any]
            Dictionary of parameter names/values

        """
        try:
            if self.primary_host:
                host_name = self.primary_host.fqdn
            else:
                host_name = None
            proc_name = (
                self.primary_process.ImageFile.FullPath
                if self.primary_process and self.primary_process.ImageFile
                else None
            )
            acct_name = self.primary_account.Name if self.primary_account else None
            path_separator = self.path_separator
            if self.data_family == DataFamily.WindowsSecurity:
                proc_name = escape_windows_path(proc_name)
                path_separator = escape_windows_path(self.path_separator)

            dyn_query_params = {
                "subscription_filter": self.subscription_filter(),
                "host_filter_eq": self.host_filter(operator="=="),
                "host_filter_neq": self.host_filter(operator="!="),
                "host_name": host_name,
                "account_name": acct_name,
                "process_name": proc_name,
                "logon_session_id": self.get_logon_id(),
                "process_id": (
                    self.primary_process.ProcessId if self.primary_process else None
                ),
                "path_separator": path_separator,
                "data_family": self.data_family,
                "data_environment": self.data_environment,
            }

            dyn_query_params.update(self._custom_query_params)
            return dyn_query_params
        except AttributeError:
            return {}

    @property
    def data_family(self) -> DataFamily:
        """Return the data family of the alert for subsequent queries."""
        if self.os_family == "Linux":
            return DataFamily.LinuxSecurity
        if self.os_family == "Windows":
            return DataFamily.WindowsSecurity
        raise ValueError("Unknown Data family.")

    @property
    def data_environment(self) -> DataEnvironment:
        """Return the data environment of the alert for subsequent queries."""
        if self.is_in_log_analytics:
            return DataEnvironment.LogAnalytics
        return DataEnvironment.Kusto

    @property
    def origin_time(self) -> datetime:
        """Return the datetime of event."""
        return self.TimeGenerated

    def get_logon_id(self, account: Account = None) -> Optional[Union[str, int]]:
        """
        Get the logon Id for the alert or the account, if supplied.

        If `account` is not supplied, return the logon id
        of the first host-logon-session or account entity.

        Parameters
        ----------
        account : Account, optional
            Account objec to use (the default is None)

        Returns
        -------
        Optional[Union[str, int]]
            The logon Id for primary account

        """
        for session in [
            e
            for e in self.entities
            if e["Type"] == "host-logon-session" or e["Type"] == "hostlogonsession"
        ]:
            if account is None or session["Account"] == account:
                return session["SessionId"]
        if account is None:
            for acct in [
                e for e in self.entities if e["Type"] == "account" and "LogonId" in e
            ]:
                return acct["LogonId"]
        elif "LogonId" in account:
            return account["LogonId"]
        return None

    def subscription_filter(self, operator="=="):
        """Return a KQL subscription filter clause derived from the alert properties."""
        if self.is_in_log_analytics:
            return "true"
        if self.is_in_azure_sub:
            return "AzureResourceSubscriptionId {} '{}'".format(
                operator, self._ids["AzSubscriptionId"]
            )
        if self.is_in_workspace:
            return "WorkspaceId {} '{}'".format(operator, self._ids["WorkspaceId"])

        # Otherwise we default to including everything
        return "true"

    def host_filter(self, operator="=="):
        """
        Return a KQL host filter clause derived from the alert properties.

            :param operator='==': the operator to use in the filter clause.
                '==' and '!=' typically.
        """
        if self.primary_host:
            case_insens_op = "=~" if operator == "==" else "!~"
            return "Computer {} '{}'".format(case_insens_op, self.primary_host.computer)

        if (
            self.is_in_log_analytics
            and "SourceComputerId" in self._ids
            and self._ids["SourceComputerId"]
        ):
            return "SourceComputerId {} '{}'".format(
                operator, self._ids["SourceComputerId"]
            )
        if (
            self.is_in_azure_sub
            and "AzureResourceId" in self._ids
            and self._ids["AzResourceId"]
        ):
            return "AzureResourceId {} '{}'".format(operator, self._ids["AzResourceId"])
        if self.is_in_workspace and "AgendId" in self._ids and self._ids["AgentId"]:
            return "AgentId {} '{}'".format(operator, self._ids["AgentId"])
        return None

    def get_entities_of_type(self, entity_type: str) -> List[Entity]:
        """
        Return entity collection for a give entity type.

        Parameters
        ----------
        entity_type : str, optional
            The entity type.

        Returns
        -------
        List[Entity]
            The entities matching `entity_type`.

        """
        class_type = Entity.ENTITY_NAME_MAP.get(entity_type, None)
        return [
            p
            for p in self.entities
            if p["Type"] == entity_type or class_type and isinstance(p, class_type)
        ]

    def get_all_entities(self) -> pd.DataFrame:
        """
        Return a DataFrame of the Alert or Event entities.

        Returns
        -------
        DataFrame
            Pandas DataFrame of the Alert or Event entities.

        """
        entity = []
        ent_type = []
        for item in self.entities:
            if "Address" in item:
                entity.append(item["Address"])
                ent_type.append(item["Type"])
            elif "Url" in item:
                entity.append(item["Url"])
                ent_type.append(item["Type"])
            elif "HostName" in item:
                entity.append(item["HostName"])
                ent_type.append(item["Type"])
            elif "Entity" in item:
                entity.append(item["Entity"])
                ent_type.append(item["Type"])
            elif item["Type"] == "account":
                entity.append(item["Name"])
                ent_type.append(item["Type"])

        entities = pd.DataFrame({"Entity": entity, "Type": ent_type})
        return entities

    def to_html(self, show_entities: bool = False) -> str:
        """Return the item as HTML string."""
        html_doc = pd.DataFrame(self._source_data).to_html()

        if self._source_data is not None and "ExtendedProperties" in self._source_data:
            ext_prop_title = "<br/><h3>ExtendedProperties:</h3>"
            ext_prop_html = pd.DataFrame(
                pd.Series(self._source_data["ExtendedProperties"])
            ).to_html()
            html_doc = html_doc + ext_prop_title + ext_prop_html

        if show_entities and self.entities:
            entity_title = "<br/><h3>Entities:</h3><br/>"
            entity_html = "<br/>".join(
                [self._format_entity(ent) for ent in self.entities]
            )
            html_doc = html_doc + entity_title + entity_html
        else:
            e_counts = Counter([ent["Type"] for ent in self.entities])
            e_counts_str = ", ".join([f"{e}: {c}" for e, c in e_counts.items()])
            html_doc = html_doc + f"<h3>Entity counts: </h3>{e_counts_str}"
        return html_doc

    @staticmethod
    def _format_entity(entity):
        str_entity = str(entity)
        if str_entity:
            str_entity = str_entity.replace("\n", ", ")
        return html.escape(str_entity)

    def _find_os_family(self):
        """Return OSFamily and path separator to use from entities or file paths."""
        self.path_separator = "\\"
        self.os_family = "Windows"

        # Use OSFamily if any entities have this property set
        os_family_entities = [e for e in self.entities if "OSFamily" in e]
        if os_family_entities:
            for os_entity in os_family_entities:
                if os_entity["OSFamily"] == "Linux":
                    self.os_family = "Linux"
                    self.path_separator = "/"
                    break
        else:
            # Otherwise try to infer from the file paths
            files = [e for e in self.entities if e["Type"] == "file"]
            if files:
                for file in files:
                    if "Directory" in file and "/" in file["Directory"]:
                        self.os_family = "Linux"
                        self.path_separator = "/"
                        break
            else:
                for proc in [
                    e
                    for e in self.entities
                    if e["Type"] == "process" and "ImageFile" in e
                ]:
                    file = proc["ImageFile"]
                    if "Directory" in file and "/" in file["Directory"]:
                        self.os_family = "Linux"
                        self.path_separator = "/"
                        break

    @staticmethod
    def _get_subscription_from_resource(resource_id) -> Optional[str]:
        """Extract subscription Id from resource string."""
        sub_regex = r"^/subscriptions/([^/]+)/"
        sub_ids = re.findall(sub_regex, resource_id, re.RegexFlag.I)
        if sub_ids:
            return sub_ids[0]

        return None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data driver base class."""
import abc
from abc import ABC
from typing import Any, Callable, Dict, Iterable, Optional, Tuple, Union

import pandas as pd

from ..._version import VERSION
from ..query_source import QuerySource

__version__ = VERSION
__author__ = "Ian Hellen"


class DriverBase(ABC):
    """Base class for data providers."""

    def __init__(self, **kwargs):
        """Initialize new instance."""
        self._kwargs = kwargs
        self._loaded = False
        self._connected = False
        self.current_connection = None
        self.public_attribs: Dict[str, Any] = {}
        self.formatters: Dict[str, Callable] = {}
        self.use_query_paths = True
        self.has_driver_queries = False

    @property
    def loaded(self) -> bool:
        """
        Return true if the provider is loaded.

        Returns
        -------
        bool
            True if the provider is loaded.

        Notes
        -----
        This is not relevant for some providers.

        """
        return self._loaded

    @property
    def connected(self) -> bool:
        """
        Return true if at least one connection has been made.

        Returns
        -------
        bool
            True if a successful connection has been made.

        Notes
        -----
        This does not guarantee that the last data source
        connection was successful. It is a best effort to track
        whether the provider has made at least one successful
        authentication.

        """
        return self._connected

    @property
    def schema(self) -> Dict[str, Dict]:
        """
        Return current data schema of connection.

        Returns
        -------
        Dict[str, Dict]
            Data schema of current connection.

        """
        return {}

    @abc.abstractmethod
    def connect(self, connection_str: Optional[str] = None, **kwargs):
        """
        Connect to data source.

        Parameters
        ----------
        connection_str : Optional[str]
            Connect to a data source

        """
        return None

    @abc.abstractmethod
    def query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Other Parameters
        ----------------
        kwargs :
            Are passed to the underlying provider query method,
            if supported.

        Returns
        -------
        Union[pd.DataFrame, Any]
            A DataFrame (if successfull) or
            the underlying provider result if an error.

        """

    @abc.abstractmethod
    def query_with_results(self, query: str, **kwargs) -> Tuple[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame plus native results.

        Parameters
        ----------
        query : str
            The query to execute

        Returns
        -------
        Tuple[pd.DataFrame,Any]
            A DataFrame and native results.

        """

    @property
    def service_queries(self) -> Tuple[Dict[str, str], str]:
        """
        Return queries retrieved from the service after connecting.

        Returns
        -------
        Tuple[Dict[str, str], str]
            Dictionary of query_name, query_text.
            Name of container to add queries to.

        """
        return {}, ""

    @property
    def driver_queries(self) -> Iterable[Dict[str, Any]]:
        """
        Return queries retrieved from the service after connecting.

        Returns
        -------
        List[Dict[str, str]]
            List of Dictionary of query_name, query_text.
            Name of container to add queries to.

        """
        return [{}]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Useful helper data structure classes for modelling sessions."""

from collections import defaultdict
from typing import Union

from ....common.exceptions import MsticpyException


class StateMatrix(dict):
    """Class for storing trained counts/probabilities."""

    def __init__(self, states: Union[dict, defaultdict], unk_token: str):
        """
        Containr for dict of counts/probs or dict of dicts of cond counts/probs.

        If you try and retrieve the count/probability for an unseen
        command/param/value from the resulting object, it will return
        the value associated with the `unk_token` key.

        Parameters
        ----------
        states: Union[dict, defaultdict]
            Either a dict representing counts or probabilities.
            Or a dict of dicts representing
            conditional counts or conditional probabilities.
            E.g.
                {'Set-Mailbox': 20,'##UNK##': 1}
            or
                {'Set-Mailbox': {'Set-Mailbox': 5, '##UNK##': 1},
                '##UNK##': {'Set-Mailbox': 1, '##UNK##': 1}}
        unk_token: str
            dummy token to signify an unseen command (e.g. "##UNK##").
            This token should be present in the `states` keys. And if `states`
            is a dict of dicts, then the `unk_token` should be present in
            the keys of the outer dict and all the inner dicts.

        """
        super().__init__(states)
        if unk_token not in states:
            raise MsticpyException("`unk_token` should be a key in `states`")
        self.states = dict(states)
        self.unk_token = unk_token
        for key, val in self.states.items():
            if isinstance(val, dict):
                self.states[key] = StateMatrix(self.states[key], unk_token)

    def __getitem__(self, item):
        """
        Get value or self.unk_token for unseen cmds/params/values.

        Parameters
        ----------
        item:
            desired key to retrieve value for from the dictionary

        Returns
        -------
        Value associated with the key `item` if it exists, else the
        value associated with the `unk_token`

        """
        if item not in self.states:
            return self.states[self.unk_token]
        return self.states[item]


class Cmd:
    """Class to store commands with accompanying params (and optionally values)."""

    def __init__(self, name: str, params: Union[set, dict]):
        """
        Instantiate the Cmd class.

        Parameters
        ----------
        name: str
            name of the command. e.g. for Exchange online: "Set-Mailbox"
        params: Union[set, dict]
            set of accompanying params or dict of accompanying params and values.
            e.g.
                {'Identity', 'ForwardingEmailAddress'}
            of
                {'Identity': 'some identity', 'ForwardingEmailAddress':
                 'an_email@email.com'}

        """
        self.name = name
        self.params = params

    def __str__(self):
        """
        Make string representation more pleasing to the eye.

        Returns
        -------
        New string representation of the object

        """
        rep = "Cmd(name='{}', params={})".format(self.name, self.params)
        return rep

    def __repr__(self):
        """
        Make object representation more pleasing to the eye.

        Returns
        -------
        New string representation of the object

        """
        return str(self)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Geoip Lookup module using IPStack and Maxmind GeoLite2.

Geographic location lookup for IP addresses. This module has two classes
for different services:

-  GeoLiteLookup - Maxmind Geolite (see https://www.maxmind.com)
-  IPStackLookup - IPStack (see https://ipstack.com)

Both services offer
a free tier for non-commercial use. However, a paid tier will
normally get you more accuracy, more detail and a higher throughput
rate. Maxmind geolite uses a downloadable database, while IPStack is
an online lookup (API key required).

"""
import math
import os
import tarfile
import warnings
from abc import ABCMeta, abstractmethod
from collections.abc import Iterable
from datetime import datetime, timedelta
from json import JSONDecodeError
from pathlib import Path
from typing import Any, Dict, List, Mapping, Optional, Tuple

import geoip2.database  # type: ignore
import pandas as pd
import requests
from geoip2.errors import AddressNotFoundError  # type: ignore
from IPython import get_ipython
from IPython.display import HTML, display
from requests.exceptions import HTTPError

from .._version import VERSION
from ..nbtools.entityschema import GeoLocation, IpAddress  # type: ignore
from ..common.utility import export
from ..common.exceptions import MsticpyUserConfigError
from ..common.provider_settings import ProviderSettings, get_provider_settings

__version__ = VERSION
__author__ = "Ian Hellen"


class GeoIPDatabaseException(Exception):
    """Exception when GeoIP database cannot be found."""


class GeoIpLookup(metaclass=ABCMeta):
    """
    Abstract base class for GeoIP Lookup classes.

    See Also
    --------
    IPStackLookup : IPStack GeoIP Implementation
    GeoLiteLookup : MaxMind GeoIP Implementation

    """

    _LICENSE_TXT: Optional[str] = None
    _LICENSE_HTML: Optional[str] = None
    _license_shown: bool = False

    def __init__(self):
        """Initialize instance of GeoIpLookup class."""
        self._print_license()

    @abstractmethod
    def lookup_ip(
        self,
        ip_address: str = None,
        ip_addr_list: Iterable = None,
        ip_entity: IpAddress = None,
    ) -> Tuple[List[Any], List[IpAddress]]:
        """
        Lookup IP location abstract method.

        Parameters
        ----------
        ip_address : str, optional
            a single address to look up (the default is None)
        ip_addr_list : Iterable, optional
            a collection of addresses to lookup (the default is None)
        ip_entity : IpAddress, optional
            an IpAddress entity (the default is None) - any existing
            data in the Location property will be overwritten

        Returns
        -------
        Tuple[List[Any], List[IpAddress]]:
            raw geolocation results and same results as IpAddress entities with
            populated Location property.

        """

    def df_lookup_ip(self, data: pd.DataFrame, column: str) -> pd.DataFrame:
        """
        Lookup Geolocation data from a pandas Dataframe.

        Parameters
        ----------
        data : pd.DataFrame
            pandas dataframe containing IpAddress column
        column : str
            the name of the dataframe column to use as a source

        Returns
        -------
        pd.DataFrame
            Copy of original dataframe with IP Location information columns
            appended (where a location lookup was successful)

        """
        ip_list = list(data[column].values)
        _, entities = self.lookup_ip(ip_addr_list=ip_list)

        ip_dicts = [
            {**ent.Location.properties, "IpAddress": ent.Address} for ent in entities
        ]
        df_out = pd.DataFrame(data=ip_dicts)
        return data.merge(df_out, how="left", left_on=column, right_on="IpAddress")

    # pylint: disable=protected-access
    def _print_license(self):
        if self.__class__._license_shown:
            return
        if self._LICENSE_HTML and get_ipython():
            display(HTML(self._LICENSE_HTML))
        elif self._LICENSE_TXT:
            print(self._LICENSE_TXT)
        self.__class__._license_shown = True

    # pylint: enable=protected-access


@export
class IPStackLookup(GeoIpLookup):
    """
    IPStack GeoIP Implementation.

    See Also
    --------
    GeoIpLookup : Abstract base class
    GeoLiteLookup : MaxMind GeoIP Implementation

    """

    _LICENSE_HTML = """
This library uses services provided by ipstack.
<a href="https://ipstack.com">https://ipstack.com</a>"""

    _LICENSE_TXT = """
This library uses services provided by ipstack (https://ipstack.com)"""

    _IPSTACK_API = "http://api.ipstack.com/{iplist}?access_key={access_key}&output=json"

    _NO_API_KEY_MSSG = """
No API Key was found to access the IPStack service.
If you do not have an account, go here to create one and obtain and API key.

Add this API key to your msticpyconfig.yaml
Alternatively, you can pass this to the IPStackLookup class when creating it:
>>> iplookup = IPStackLookup(api_key="your_api_key")
"""

    def __init__(self, api_key: Optional[str] = None, bulk_lookup: bool = False):
        """
        Create a new instance of IPStackLookup.

        Parameters
        ----------
        api_key : str, optional
            API Key from IPStack - see https://ipstack.com
            default is None - obtain key from msticpyconfig.yaml
        bulk_lookup : bool, optional
            For Professional and above tiers allowing you to
            submit multiple IPs in a single request.
            (the default is False, which submits a single request
            per address)

        """
        super().__init__()

        self.settings = _get_geoip_provider_settings("IPStack")
        if api_key:
            self._api_key = api_key
        else:
            self._api_key = self.settings.args.get("AuthKey")  # type: ignore
        if not self._api_key:
            raise MsticpyUserConfigError(
                self._NO_API_KEY_MSSG,
                help_uri=(
                    "https://msticpy.readthedocs.io/en/latest/data_acquisition/"
                    + "GeoIPLookups.html#ipstack-geo-lookup-class"
                ),
                service_uri="https://ipstack.com/product",
                title="IPStack API key not found",
            )
        self.bulk_lookup = bulk_lookup

    def lookup_ip(
        self,
        ip_address: str = None,
        ip_addr_list: Iterable = None,
        ip_entity: IpAddress = None,
    ) -> Tuple[List[Any], List[IpAddress]]:
        """
        Lookup IP location from IPStack web service.

        Parameters
        ----------
        ip_address : str, optional
            a single address to look up (the default is None)
        ip_addr_list : Iterable, optional
            a collection of addresses to lookup (the default is None)
        ip_entity : IpAddress, optional
            an IpAddress entity (the default is None) - any existing
            data in the Location property will be overwritten

        Returns
        -------
        Tuple[List[Any], List[IpAddress]]:
            raw geolocation results and same results as IpAddress entities with
            populated Location property.

        Raises
        ------
        ConnectionError
            Invalid status returned from http request
        PermissionError
            Service refused request (e.g. requesting batch of addresses
            on free tier API key)

        """
        if ip_address and isinstance(ip_address, str):
            ip_list = [ip_address.strip()]
        elif ip_addr_list:
            ip_list = list((ip.strip() for ip in ip_addr_list))
        elif ip_entity:
            ip_list = [ip_entity.Address]
        else:
            raise ValueError("No valid ip addresses were passed as arguments.")

        results = self._submit_request(ip_list)
        output_raw = []
        output_entities = []
        for ip_loc, status in results:
            if status == 200:
                output_entities.append(self._create_ip_entity(ip_loc, ip_entity))
            output_raw.append((ip_loc, status))
        return output_raw, output_entities

    @staticmethod
    def _create_ip_entity(ip_loc: dict, ip_entity) -> IpAddress:
        if not ip_entity:
            ip_entity = IpAddress()
            ip_entity.Address = ip_loc["ip"]
        geo_entity = GeoLocation()
        geo_entity.CountryCode = ip_loc["country_code"]  # type: ignore

        geo_entity.CountryName = ip_loc["country_name"]  # type: ignore
        geo_entity.State = ip_loc["region_name"]  # type: ignore
        geo_entity.City = ip_loc["city"]  # type: ignore
        geo_entity.Longitude = ip_loc["longitude"]  # type: ignore
        geo_entity.Latitude = ip_loc["latitude"]  # type: ignore
        if "connection" in ip_loc:
            geo_entity.Asn = ip_loc["connection"]["asn"]  # type: ignore
        ip_entity.Location = geo_entity
        return ip_entity

    def _submit_request(self, ip_list: List[str]) -> List[Tuple[Dict[str, str], int]]:
        """
        Submit the request to IPStack.

        Parameters
        ----------
        ip_list : List[str]
            String list of IPs to look up

        Returns
        -------
        List[Tuple[str, int]]
            List of response, status code pairs

        """
        if not self.bulk_lookup:
            return self._lookup_ip_list(ip_list)

        submit_url = self._IPSTACK_API.format(
            iplist=",".join(ip_list), access_key=self._api_key
        )
        response = requests.get(submit_url)

        if response.status_code == 200:
            results = response.json()
            # {"success":false,"error":{"code":303,"type":"batch_not_supported_on_plan",
            # "info":"Bulk requests are not supported on your plan.
            # Please upgrade your subscription."}}

            if "success" in results and not results["success"]:
                raise PermissionError(
                    "Service unable to complete request. Error: {}".format(
                        results["error"]
                    )
                )
            return [(item, response.status_code) for item in results]

        if response:
            try:
                return [(response.json(), response.status_code)]
            except JSONDecodeError:
                pass
        return [({}, response.status_code)]

    def _lookup_ip_list(self, ip_list: List[str]):
        """Lookup IP Addresses one-by-one."""
        ip_loc_results = []
        with requests.Session() as session:
            for ip_addr in ip_list:
                submit_url = self._IPSTACK_API.format(
                    iplist=ip_addr, access_key=self._api_key
                )
                response = session.get(submit_url)
                if response.status_code == 200:
                    ip_loc_results.append((response.json(), response.status_code))
                else:
                    if response:
                        try:
                            ip_loc_results.append(
                                (response.json(), response.status_code)
                            )
                            continue
                        except JSONDecodeError:
                            ip_loc_results.append((None, response.status_code))
                    else:
                        print("Unknown response from IPStack request.")
                        ip_loc_results.append((None, -1))
        return ip_loc_results


@export
class GeoLiteLookup(GeoIpLookup):
    """
    GeoIP Lookup using MaxMindDB database.

    See Also
    --------
    GeoIpLookup : Abstract base class
    IPStackLookup : IPStack GeoIP Implementation

    """

    _MAXMIND_DOWNLOAD = (
        "https://download.maxmind.com/app/geoip_download?"
        + "edition_id=GeoLite2-City&license_key={license_key}&suffix=tar.gz"
    )

    _DB_HOME = os.path.join(os.path.expanduser("~"), ".msticpy", "GeoLite2")
    _DB_ARCHIVE = "GeoLite2-City.mmdb.tar.gz"
    _DB_FILE = "GeoLite2-City.mmdb"

    _LICENSE_HTML = """
This product includes GeoLite2 data created by MaxMind, available from
<a href="https://www.maxmind.com">https://www.maxmind.com</a>.
"""

    _LICENSE_TXT = """
This product includes GeoLite2 data created by MaxMind, available from
https://www.maxmind.com.
"""

    _NO_API_KEY_MSSG = """
No API Key was found to download the Maxmind GeoIPLite database.
If you do not have an account, go here to create one and obtain and API key.
https://www.maxmind.com/en/geolite2/signup

Add this API key to your msticpyconfig.yaml
https://msticpy.readthedocs.io/en/latest/data_acquisition/GeoIPLookups.html#maxmind-geo-ip-lite-lookup-class.
Alternatively, you can pass this to the GeoLiteLookup class when creating it:
>>> iplookup = GeoLiteLookup(api_key="your_api_key")
"""

    def __init__(
        self,
        api_key: Optional[str] = None,
        db_folder: Optional[str] = None,
        force_update: bool = False,
        auto_update: bool = True,
    ):
        r"""
        Return new instance of GeoLiteLookup class.

        Parameters
        ----------
        api_key : str, optional
            Default is None - use configuration value from msticpyconfig.yaml.
            API Key from MaxMind -
            Read more about GeoLite2 : https://dev.maxmind.com/geoip/geoip2/geolite2/
            Sign up for a MaxMind account:
            https://www.maxmind.com/en/geolite2/signup
            Set your password and create a license key:
            https://www.maxmind.com/en/accounts/current/license-key
        db_folder: str, optional
            Provide absolute path to the folder containing MMDB file
            (e.g. '/usr/home' or 'C:/maxmind').
            If no path provided, it is set to download to .msticpy/GeoLite2
            under user`s home directory.
        force_update : bool, optional
            Force update can be set to true or false. depending on it,
            new download request will be initiated.
        auto_update: bool, optional
            Auto update can be set to true or false. depending on it,
            new download request will be initiated if age criteria is matched.

        """
        super().__init__()

        self.settings = _get_geoip_provider_settings("GeoIPLite")
        if api_key:
            self._api_key = api_key
        else:
            self._api_key = self.settings.args.get("AuthKey")  # type: ignore

        self._dbfolder = db_folder
        if self._dbfolder is None:
            self._dbfolder = self.settings.args.get("DBFolder", self._DB_HOME)

        self._dbfolder = str(Path(self._dbfolder).expanduser())  # type: ignore
        self._force_update = force_update
        self._auto_update = auto_update
        self._check_and_update_db(self._dbfolder, self._force_update, self._auto_update)
        self._dbpath = self._get_geoip_dbpath(self._dbfolder)
        if not self._dbpath:
            raise MsticpyUserConfigError(
                "No usable GeoIP Database could be found.",
                "Check that you have correctly configured the Maxmind API key.",
                (
                    "If you are using a custom DBFolder setting in your config, "
                    + "check that this is a valid path."
                ),
                help_uri=(
                    "https://msticpy.readthedocs.io/en/latest/data_acquisition/"
                    + "GeoIPLookups.html#maxmind-geo-ip-lite-lookup-class"
                ),
                service_uri="https://www.maxmind.com/en/geolite2/signup",
                title="Maxmind GeoIP database not found",
            )
        self._reader = geoip2.database.Reader(self._dbpath)

    def _check_and_update_db(
        self,
        db_folder: str = None,
        force_update: bool = False,
        auto_update: bool = True,
    ):
        r"""
        Check the age of geo ip database file and download if it older than 30 days.

        User can set auto_update or force_update to True or False to
        override auto-download behavior.

        Parameters
        ----------
        db_folder: str, optional
            Provide absolute path to the folder containing MMDB file
            (e.g. '/usr/home' or 'C:\maxmind').
            If no path provided, it is set to download to .msticpy\GeoLite2 dir under
            user`s home directory.
        force_update : bool, optional
            Force update can be set to true or false. depending on it,
            new download request will be initiated, overriding age criteria.
        auto_update : bool, optional
            Auto update can be set to true or false. depending on it,
            new download request will be initiated if age criteria is matched.

        """
        geoip_db_path = self._get_geoip_dbpath(db_folder)
        url = self._MAXMIND_DOWNLOAD.format(license_key=self._api_key)

        if geoip_db_path is None:
            print(
                "No local Maxmind City Database found. ",
                f"Attempting to downloading new database to {db_folder}",
            )
            self._download_and_extract_archive(url, db_folder)
        else:
            # Create a reader object to retrive db info and build date
            # to check age from build_epoch property.
            with geoip2.database.Reader(geoip_db_path) as reader:
                last_mod_time = datetime.utcfromtimestamp(reader.metadata().build_epoch)

            # Check for out of date DB file according to db_age
            db_age = datetime.utcnow() - last_mod_time
            db_updated = True
            if db_age > timedelta(30) and auto_update:
                print(
                    "Latest local Maxmind City Database present is older than 30 days.",
                    f"Attempting to download new database to {db_folder}",
                )
                if not self._download_and_extract_archive(url, db_folder):
                    warnings.warn("DB download failed")
                    db_updated = False
            elif force_update:
                print(
                    "force_update is set to True.",
                    f"Attempting to download new database to {db_folder}",
                )
                if not self._download_and_extract_archive(url, db_folder):
                    warnings.warn("DB download failed")
                    db_updated = False
            if not db_updated:
                warnings.warn(
                    "Continuing with cached database. Results may inaccurate."
                )

    def _download_and_extract_archive(  # noqa: MC0001
        self, url: str = None, db_folder: str = None
    ) -> bool:
        r"""
        Download file from the given URL and extract if it is archive.

        Parameters
        ----------
        url : str
            Web URL location to the Maxmind city Database. (the default is None)
        db_folder: str, optional
            Provide absolute path to the folder containing MMDB file
            (e.g. '/usr/home' or 'C:\maxmind').
            If no path provided, it is set to download to .msticpy dir under
            user`s home directory.(the default is None)

        Returns
        -------
        bool :
            True if download_success

        """
        if not self._api_key:
            return False
        if url is None:
            url = self._MAXMIND_DOWNLOAD.format(license_key=self._api_key)

        if db_folder is None:
            db_folder = self._DB_HOME

        if not os.path.exists(db_folder):
            # using makedirs to create intermediate-level dirs to contain the leaf dir
            os.makedirs(db_folder)
        db_archive_path = os.path.join(db_folder, self._DB_ARCHIVE)
        db_file_path = os.path.join(db_folder, self._DB_FILE)

        try:
            with requests.get(url, stream=True) as response:
                response = requests.get(url, stream=True)
                response.raise_for_status()
                print("Downloading and extracting GeoLite DB archive from MaxMind....")
                with open(db_archive_path, "wb") as file_hdl:
                    for chunk in response.iter_content(chunk_size=10000):
                        file_hdl.write(chunk)
                        file_hdl.flush()
        except HTTPError as http_err:
            warnings.warn(
                f"HTTP error occurred trying to download GeoLite DB: {http_err}"
            )
        # pylint: disable=broad-except
        except Exception as err:
            warnings.warn(f"Other error occurred trying to download GeoLite DB: {err}")
        # pylint: enable=broad-except
        else:
            try:
                tar_archive = tarfile.open(db_archive_path)
                for member in tar_archive.getmembers():
                    if (
                        member.isreg()
                    ):  # Will skip the dirs to extract only file objects
                        # Strip the path from files to extract it to desired directory
                        member.name = os.path.basename(member.name)
                        tar_archive.extract(member, db_folder)
                print("Extraction complete. Local Maxmind city DB:", f"{db_file_path}")
                return True
            except IOError as err:
                warnings.warn(f"Error writing GeoIP DB file: {db_file_path} - {err}")
        return False

    @staticmethod
    def _get_geoip_dbpath(db_folder: str = None) -> Optional[str]:
        r"""
        Get the correct path containing GeoLite City Database.

        Parameters
        ----------
        db_folder: str, optional
            Provide absolute path to the folder containing MMDB file
            (e.g. '/usr/home' or 'C:\maxmind').
            If no path provided, it is set to download to .msticpy\GeoLite2 dir under
            user`s home directory.

        Returns
        -------
        Optional[str]
            Returns the absolute path of local maxmind geolite city
            database after control flow logic.

        """
        if not db_folder:
            db_folder = "."
        list_of_db_paths = [str(db) for db in Path(db_folder).glob("*.mmdb")]

        if len(list_of_db_paths) > 1:
            latest_db_path = max(list_of_db_paths, key=os.path.getmtime)
        elif len(list_of_db_paths) == 1:
            latest_db_path = list_of_db_paths[0]
        else:
            return None

        return latest_db_path

    def lookup_ip(
        self,
        ip_address: str = None,
        ip_addr_list: Iterable = None,
        ip_entity: IpAddress = None,
    ) -> Tuple[List[Any], List[IpAddress]]:
        """
        Lookup IP location from GeoLite2 data created by MaxMind.

        Parameters
        ----------
        ip_address : str, optional
            a single address to look up (the default is None)
        ip_addr_list : Iterable, optional
            a collection of addresses to lookup (the default is None)
        ip_entity : IpAddress, optional
            an IpAddress entity (the default is None) - any existing
            data in the Location property will be overwritten

        Returns
        -------
        Tuple[List[Any], List[IpAddress]]
            raw geolocation results and same results as IpAddress entities with
            populated Location property.

        """
        if ip_address and isinstance(ip_address, str):
            ip_list = [ip_address.strip()]
        elif ip_addr_list:
            ip_list = list((ip.strip() for ip in ip_addr_list))
        elif ip_entity:
            ip_list = [ip_entity.Address]
        else:
            raise ValueError("No valid ip addresses were passed as arguments.")

        output_raw = []
        output_entities = []
        for ip_input in ip_list:
            geo_match = None
            try:
                geo_match = self._reader.city(ip_input).raw
            except (AddressNotFoundError, AttributeError, ValueError):
                continue
            if geo_match:
                output_raw.append(geo_match)
                output_entities.append(
                    self._create_ip_entity(ip_input, geo_match, ip_entity)
                )

        return output_raw, output_entities

    @staticmethod
    def _create_ip_entity(
        ip_address: str, geo_match: Mapping[str, Any], ip_entity: IpAddress = None
    ) -> IpAddress:
        if not ip_entity:
            ip_entity = IpAddress()
            ip_entity.Address = ip_address
        geo_entity = GeoLocation()
        geo_entity.CountryCode = geo_match.get("country", {}).get(  # type: ignore
            "iso_code", None
        )
        geo_entity.CountryName = (  # type: ignore
            geo_match.get("country", {}).get("names", {}).get("en", None)
        )
        subdivs = geo_match.get("subdivisions", [])
        if subdivs:
            geo_entity.State = (  # type: ignore
                subdivs[0].get("names", {}).get("en", None)
            )
        geo_entity.City = (  # type: ignore
            geo_match.get("city", {}).get("names", {}).get("en", None)
        )
        geo_entity.Longitude = geo_match.get("location", {}).get(  # type: ignore
            "longitude", None
        )
        geo_entity.Latitude = geo_match.get("location", {}).get(  # type: ignore
            "latitude", None
        )
        ip_entity.Location = geo_entity  # type: ignore
        return ip_entity


def _get_geoip_provider_settings(provider_name: str) -> ProviderSettings:
    """
    Return settings for a provider.

    Parameters
    ----------
    provider_name : str
        Name of the provider.

    Returns
    -------
    ProviderSettings
        Settings for the provider.

    """
    settings = get_provider_settings(config_section="OtherProviders")
    if provider_name in settings:
        return settings[provider_name]
    return ProviderSettings(name=provider_name, description="Not found.")


@export
def entity_distance(ip_src: IpAddress, ip_dest: IpAddress) -> float:
    """
    Return distance between two IP Entities.

    Parameters
    ----------
    ip_src : IpAddress
        Source/Origin IpAddress Entity
    ip_dest : IpAddress
        Destination IpAddress Entity

    Returns
    -------
    float
        Distance in kilometers.

    Raises
    ------
        AttributeError
            If either entity has no location information

    """
    if not ip_src.Location or not ip_dest.Location:
        raise AttributeError(
            "Source and destination entities must have defined Location properties."
        )

    return geo_distance(
        origin=(ip_src.Location.Latitude, ip_src.Location.Longitude),
        destination=(ip_dest.Location.Latitude, ip_dest.Location.Longitude),
    )


_EARTH_RADIUS_KM = 6371  # km


@export
def geo_distance(
    origin: Tuple[float, float], destination: Tuple[float, float]
) -> float:
    """
    Calculate the Haversine distance.

    Parameters
    ----------
    origin : Tuple[float, float]
        Latitude, Longitude of origin of distance measurement.
    destination : Tuple[float, float]
        Latitude, Longitude of origin of distance measurement.

    Returns
    -------
    float
        Distance in kilometers.

    Examples
    --------
    >>> origin = (48.1372, 11.5756)  # Munich
    >>> destination = (52.5186, 13.4083)  # Berlin
    >>> round(geo_distance(origin, destination), 1)
    504.2

    Notes
    -----
        Author: Martin Thoma - stackoverflow

    """
    orig_lat, orig_lon = origin
    dest_lat, dest_lon = destination

    ang_dist_lat = math.radians(dest_lat - orig_lat)
    ang_dist_lon = math.radians(dest_lon - orig_lon)
    hav_a = (math.sin(ang_dist_lat / 2) * math.sin(ang_dist_lat / 2)) + (
        math.cos(math.radians(orig_lat))
        * math.cos(math.radians(dest_lat))
        * math.sin(ang_dist_lon / 2)
        * math.sin(ang_dist_lon / 2)
    )
    hav_c = 2 * math.atan2(math.sqrt(hav_a), math.sqrt(1 - hav_a))
    return _EARTH_RADIUS_KM * hav_c
